""" Web interface to the disco-dop parser. Requires Flask.
Expects a series of grammars produced by runexp in subdirectories of grammar/

Also usable from the command line:
$ curl http://localhost:5000/parse -G --data-urlencode "sent=What's up?"
"""
import os
import re
import glob
import heapq
import string  # pylint: disable=W0402
import random
import logging
from urllib import urlencode
from operator import itemgetter
from flask import Flask, Markup, Response
from flask import request, render_template, send_from_directory
from werkzeug.contrib.cache import SimpleCache
from discodop import treebank
from discodop.tree import Tree
from discodop.treedraw import DrawTree
from discodop.runexp import readparam
from discodop.parser import Parser, readgrammars, probstr

LIMIT = 40  # maximum sentence length
APP = Flask(__name__)
CACHE = SimpleCache()
PARSERS = {}


@APP.route('/')
def main():
	""" Serve the main form. """
	return render_template('parse.html', result=Markup(parse()))


@APP.route('/parse')
def parse():
	""" Parse sentence and return a textual representation of a parse tree,
	in a HTML fragment or plain text. To be invoked by an AJAX call."""
	sent = request.args.get('sent', None)
	est = request.args.get('est', 'dop1')
	marg = request.args.get('marg', 'nbest')
	objfun = request.args.get('objfun', 'mpp')
	coarse = request.args.get('coarse', None)
	html = request.args.get('html', False)
	if not sent:
		return ''
	frags = nbest = None
	senttok = tokenize(sent)
	if not senttok or not 1 <= len(senttok) <= LIMIT:
		return 'Sentence too long: %d words, max %d' % (len(senttok), LIMIT)
	key = (senttok, est, marg, objfun, coarse, html)
	link = urlencode(dict(sent=sent, est=est, marg=marg, objfun=objfun,
			coarse=coarse, html=html))
	if CACHE.get(key) is not None:
		return CACHE.get(key)
	lang = guesslang(senttok)
	PARSERS[lang].stages[-1].estimator = est
	PARSERS[lang].stages[-1].objective = objfun
	PARSERS[lang].stages[-1].kbest = marg in ('nbest', 'both')
	PARSERS[lang].stages[-1].sample = marg in ('sample', 'both')
	if PARSERS[lang].stages[0].mode.startswith('pcfg') and coarse:
		PARSERS[lang].stages[0].mode = coarse
		PARSERS[lang].stages[1].k = 1e-5 if coarse == 'pcfg-posterior' else 50

	results = list(PARSERS[lang].parse(senttok))
	if results[-1].noparse:
		parsetrees = {}
		result = 'no parse!'
		frags = nbest = ''
		msg = ''
	else:
		if PARSERS[lang].relationalrealizational:
			treebank.handlefunctions('add', results[-1].parsetree, pos=True)
		tree = str(results[-1].parsetree)
		prob = results[-1].prob
		parsetrees = results[-1].parsetrees or {}
		parsetrees = heapq.nlargest(10, parsetrees.items(), key=itemgetter(1))
		fragments = results[-1].fragments or ()
		msg = '\n'.join(stage.msg for stage in results)
		APP.logger.info('[%s] %s' % (probstr(prob), tree))
		tree = Tree.parse(tree, parse_leaf=int)
		result = Markup(DrawTree(tree, senttok, abbr=True).text(
				unicodelines=True, html=html))
		frags = Markup('Phrasal fragments used in the most probable derivation'
				' of the highest ranked parse tree:\n'
				+ '\n\n'.join(
				DrawTree(Tree.parse(frag, parse_leaf=int), terminals).text(
						unicodelines=True, html=html)
				for frag, terminals in fragments))
		nbest = Markup('\n\n'.join('%d. [%s]\n%s' % (n + 1, probstr(prob),
					DrawTree(PARSERS[lang].postprocess(tree)[0], senttok,
						abbr=True).text(unicodelines=True, html=html))
				for n, (tree, prob) in enumerate(parsetrees)))
	elapsed = [stage.elapsedtime for stage in results]
	elapsed = 'CPU time elapsed: %s => %gs' % (
			' '.join('%gs' % a for a in elapsed), sum(elapsed))
	info = '\n'.join(('sentence length: %d; est=%s; objfun=%s; marg=%s' % (
			len(senttok), est, objfun, marg), msg, elapsed,
			'10 most probable parse trees:',
			'\n'.join('%d. [%s] %s' % (n + 1, probstr(prob), tree)
					for n, (tree, prob) in enumerate(parsetrees)) + '\n'))
	if not html:
		CACHE.set(key, Response('\n'.join((nbest, frags, info, result)),
				mimetype='text/plain'), timeout=5000)
	CACHE.set(key, render_template('parsetree.html', sent=sent, result=result,
			frags=frags, nbest=nbest, info=info, link=link, randid=randid()),
			timeout=5000)
	return CACHE.get(key)


@APP.route('/favicon.ico')
def favicon():
	""" Serve the favicon. """
	return send_from_directory(os.path.join(APP.root_path, 'static'),
			'parse.ico', mimetype='image/vnd.microsoft.icon')


def loadparsers():
	""" Load grammars if necessary. """
	if not PARSERS:
		for directory in glob.glob('grammars/*/'):
			_, lang = os.path.split(os.path.dirname(directory))
			APP.logger.info('Loading grammar %r', lang)
			params = readparam(os.path.join(directory, 'params.prm'))
			params['resultdir'] = directory
			stages = params['stages']
			postagging = params['postagging']
			readgrammars(directory, stages, postagging,
					top=params.get('top', 'ROOT'))
			PARSERS[lang] = Parser(stages,
					transformations=params.get('transformations'),
					tailmarker=params.get('tailmarker'),
					postagging=postagging if postagging and
					postagging['method'] == 'unknownword' else None,
					relationalrealizational=params.get(
						'relationalrealizational'))
			APP.logger.info('Grammar for %s loaded.' % lang)
	assert PARSERS, 'no grammars found!'


def randid():
	""" return a string with 6 random letters. """
	return ''.join(random.choice(string.ascii_letters)
		for _ in range(6))


# List of contractions adapted from Robert MacIntyre's tokenizer.
CONTRACTIONS2 = re.compile(
		"(?i)(?:%s)\b" % "|".join([
		r"(.)('ll|'re|'ve|n't|'s|'m|'d)",
		r"\b(can)(not)",
		r"\b(D)('ye)",
		r"\b(Gim)(me)",
		r"\b(Gon)(na)",
		r"\b(Got)(ta)",
		r"\b(Lem)(me)",
		r"\b(Mor)('n)",
		r"\b(T)(is)",
		r"\b(T)(was)",
		r"\b(Wan)(na)"]))
CONTRACTIONS3 = re.compile(r"(?i)\b(?:(Whad)(dd)(ya)|(Wha)(t)(cha))\b")


def tokenize(text):
	""" Adapted from nltk.tokenize.TreebankTokenizer. """
	text = CONTRACTIONS2.sub(r'\1 \2', text)
	text = CONTRACTIONS3.sub(r'\1 \2 \3', text)
	# Separate most punctuation
	text = re.sub(r"([^\w\.\'\-\/,&])", r' \1 ', text, flags=re.UNICODE)
	# Separate commas if they're followed by space; e.g., don't separate 2,500
	# Separate single quotes if they're followed by a space.
	text = re.sub(r"([,']\s)", r' \1', text)
	# Separate periods that come before newline or end of string.
	text = re.sub(r'\. *(\n|$)', ' . ', text)
	return tuple(text.split())


def guesslang(sent):
	""" Heuristic: pick language that contains most words from input. """
	lang = max(PARSERS, key=lambda x: len(set(sent).intersection(
			PARSERS[x].stages[0].grammar.lexicalbyword)))
	APP.logger.info('Lang: %s; Sent: %s' % (lang, ' '.join(sent)))
	return lang


if __name__ == '__main__':
	logging.basicConfig()
	for log in (logging.getLogger(), APP.logger):
		log.setLevel(logging.DEBUG)
		log.handlers[0].setFormatter(logging.Formatter(
				fmt='%(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S'))
	loadparsers()
	APP.run(debug=False, host='0.0.0.0')



""" Assorted functions to read off grammars from treebanks. """
from __future__ import division, print_function
import io
import re
import sys
import codecs
import logging
from math import exp
from operator import mul, itemgetter
from fractions import Fraction
from collections import defaultdict, Counter as multiset
from itertools import count, islice, repeat
from discodop.tree import ImmutableTree, Tree
if sys.version[0] >= '3':
	from functools import reduce  # pylint: disable=W0622
	unicode = str  # pylint: disable=W0622,C0103

FORMAT = """The PLCFRS format is as follows. Rules are delimited by newlines.
Fields are separated by tabs. The fields are:

LHS	RHS1	[RHS2]	yield-function	weight

The yield function defines how the spans of the RHS nonterminals
are combined to form the spans of the LHS nonterminal. Components of the yield
function are comma-separated, 0 refers to a component of the first RHS
nonterminal, and 1 from the second. Weights are expressed as rational
fractions.
The lexicon is defined in a separate file. Lines start with a single word,
followed by pairs of possible tags and their probabilities:

WORD	TAG1	PROB1	[TAG2	PROB2 ...]

Example:
rules:   S	NP	VP	010	1/2
         VP_2	VB	NP	0,1	2/3
         NP	NN	0	1/4
lexicon: Haus	NN	3/10	JJ	1/9"""

USAGE = """Read off grammars from treebanks.
Usage: %s pcfg <input> <output> [options]
  %s plcfrs <input> <output> [options]
  %s dopreduction <input> <output> [options]
  %s doubledop <input> <output> [options]

input is a binarized treebank,
output is the base name for the filenames to write the grammar to.

Options (* marks default option):
  --inputfmt=[*export|discbracket|bracket]
  --inputenc=[*UTF-8|ISO-8859-1|...]
  --dopestimator=[*dop1|ewe|shortest|...]
  --numproc=[*1|2|...]  only relevant for double dop fragment extraction
  --gzip                compress output with gzip, view with zless &c.
  --packed              use packed graph encoding for DOP reduction

When a PCFG is requested, or the input format is 'bracket' (Penn format), the
output will be in bitpar format. Otherwise the grammar is written as a PLCFRS.
The encoding of the input treebank may be specified. Output encoding will be
ASCII for the rules, and UTF-8 for the lexicon.
\n%s\n""" % (sys.argv[0], sys.argv[0], sys.argv[0], sys.argv[0], FORMAT)


def lcfrs_productions(tree, sent, frontiers=False):
	""" Given a tree with integer indices as terminals, and a sentence
	with the corresponding words for these indices, produce a sequence
	of LCFRS productions. Always produces monotone LCFRS rules.
	For best results, tree should be canonicalized.
	When frontiers is true, frontier nodes will generate empty productions,
	by default they are ignored.

	>>> tree = Tree.parse("(S (VP_2 (V 0) (ADJ 2)) (NP 1))", parse_leaf=int)
	>>> sent = "is Mary happy".split()
	>>> lcfrs_productions(tree, sent)
	[(('S', 'VP_2', 'NP'), ((0, 1, 0),)),
	(('VP_2', 'V', 'ADJ'), ((0,), (1,))),
	(('V', 'Epsilon'), ('is',)),
	(('ADJ', 'Epsilon'), ('happy',)),
	(('NP', 'Epsilon'), ('Mary',))]
	"""
	leaves = tree.leaves()
	assert len(set(leaves)) == len(leaves), (
		"indices should be unique. indices: %r\ntree: %s" % (leaves, tree))
	assert sent, ("no sentence.\n"
		"tree: %s\nindices: %r\nsent: %r" % (tree.pprint(), leaves, sent))
	assert all(isinstance(a, int) for a in leaves), (
		"indices should be integers.\ntree: %s\nindices: %r\nsent: %r" % (
		tree.pprint(), leaves, sent))
	assert all(0 <= a < len(sent) for a in leaves), (
		"indices should point to a word in the sentence.\n"
		"tree: %s\nindices: %r\nsent: %r" % (tree.pprint(), leaves, sent))
	rules = []
	for st in tree.subtrees():
		if not st:
			raise ValueError(("Empty node. Frontier nodes should designate "
				"which part(s) of the sentence they contribute to.\ntree:"
				"%s\nindices: %r\nsent: %r" % (tree.pprint(), leaves, sent)))
		#elif all(isinstance(a, int) for a in st):
		elif isinstance(st[0], int):
			if len(st) == 1 and sent[st[0]] is not None:  # terminal node
				rule = ((st.label, 'Epsilon'), (sent[st[0]], ))
			#elif all(sent[a] is None for a in st): # frontier node
			elif frontiers:
				rule = ((st.label, ), ())
			else:
				continue
			#else:
			#	raise ValueError(("Preterminals should dominate a single "
			#		"terminal; frontier nodes should dominate a sequence of "
			#		"indices that are None in the sentence.\n"
			#		"subtree: %s\nsent: %r" % (st, sent)))
		elif all(isinstance(a, Tree) for a in st):  # isinstance(st[0], Tree):
			# convert leaves() to bitsets
			childleaves = [a.leaves() if isinstance(a, Tree) else [a]
					for a in st]
			leaves = [(idx, n) for n, child in enumerate(childleaves)
					for idx in child]
			leaves.sort(key=itemgetter(0), reverse=True)
			#tmpleaves = leaves[:]
			previdx, prevparent = leaves.pop()
			yf = [[prevparent]]
			while leaves:
				idx, parent = leaves.pop()
				if idx != previdx + 1:  # a discontinuity
					yf.append([parent])
				elif parent != prevparent:  # switch to a different non-terminal
					yf[-1].append(parent)
				# otherwise terminal is part of current range
				previdx, prevparent = idx, parent
			nonterminals = (st.label, ) + tuple(a.label for a in st)
			rule = (nonterminals, tuple(map(tuple, yf)))
		else:
			raise ValueError("Neither Tree node nor integer index:\n"
				"%r, %r" % (st[0], type(st[0])))
		rules.append(rule)
	return rules


def treebankgrammar(trees, sents):
	""" Induce a probabilistic LCFRS with relative frequencies of productions.
	When trees contain no discontinuities, the result is equivalent to a
	treebank PCFG. """
	grammar = multiset(rule for tree, sent in zip(trees, sents)
			for rule in lcfrs_productions(tree, sent))
	lhsfd = multiset()
	for rule, freq in grammar.items():
		lhsfd[rule[0][0]] += freq
	for rule, freq in grammar.items():
		grammar[rule] = Fraction(freq, lhsfd[rule[0][0]])
	return sortgrammar(grammar.items())


def dopreduction(trees, sents, packedgraph=False):
	""" Induce a reduction of DOP to an LCFRS, similar to how Goodman (1996)
	reduces DOP1 to a PCFG.

	:param packedgraph: packed graph encoding (Bansal & Klein 2010).
	:returns: a set of rules with the relative frequency estimate as
		probilities, and a dictionary with alternate weights. """
	# fd: how many subtrees are headed by node X (e.g. NP or NP@1-2),
	# 	counts of NP@... should sum to count of NP
	# ntfd: frequency of a node in treebank
	fd = defaultdict(int)
	ntfd = defaultdict(int)
	rules = defaultdict(int)
	decorater = TreeDecorator(memoize=packedgraph)

	# collect rules
	for tree, sent in zip(trees, sents):
		prods = lcfrs_productions(tree, sent)
		dectree = decorater.decorate(tree, sent)
		uprods = lcfrs_productions(dectree, sent)
		nodefreq(tree, dectree, fd, ntfd)
		for (a, avar), (b, bvar) in zip(prods, uprods):
			assert avar == bvar
			for c in cartpi([(x, ) if x == y else (x, y) for x, y in zip(a, b)]):
				rules[c, avar] += 1

	def weights(rule):
		""" :returns: rule with RFE and EWE probability. """
		# relative frequency estimate, aka DOP1 (Bod 1992; Goodman 1996, 2003)
		(r, yf), freq = rule
		rfe = Fraction((1 if '@' in r[0] else freq) *
				reduce(mul, (fd[z] for z in r[1:] if '@' in z), 1), fd[r[0]])
		# Bod (2003, figure 3): correction factor for number of subtrees.
		# Caveat: the original formula (Goodman 2003, eq. 8.23) has a_j in the
		# denominator of all rules; this is probably a misprint.
		ewe = (float((1 if '@' in r[0] else freq) *
				reduce(mul, (fd[z] for z in r[1:] if '@' in z), 1))
				/ (fd[r[0]] * (ntfd[r[0]] if '@' not in r[0] else 1)))
		# Goodman (2003, p 135). any rule corresponding to the introduction of
		# a fragment has a probability of 1/2, else 1.
		shortest = 1. if '@' in r[0] else 0.5
		# Goodman (2003, eq. 8.22). Prob. of fragment is reduced by factor of 2
		# for each non-root non-terminal it contains.
		bon = 0.25 if '@' in r[0] else (1 / (4 * ntfd[r[0]]))
		return ((r, yf), rfe), ewe, shortest, bon

	rules = sortgrammar(rules.items())
	rules, ewe, shortest, bon = zip(*(weights(r) for r in rules))
	return list(rules), dict(
			ewe=list(ewe), shortest=list(shortest), bon=list(bon))


def doubledop(trees, fragments, debug=False):
	""" Extract a Double-DOP grammar from a treebank. That is, a fragment
	grammar containing all fragments that occur at least twice, plus all
	individual productions needed to obtain full coverage.
	Input trees need to be binarized. A second level of binarization (a normal
	form) is needed when fragments are converted to individual grammar rules,
	which occurs through the removal of internal nodes. The binarization adds
	unique identifiers so that each grammar rule can be mapped back to its
	fragment. In fragments with terminals, we replace their POS tags with a tag
	uniquely identifying that terminal and tag: tag@word.

	:returns: a tuple (grammar, altweights, backtransform)
		altweights is a dictionary containing alternate weights.
	"""
	def getweight(frag, terminals):
		""" :returns: frequency, EWE, and other weights for fragment. """
		freq = sum(fragments[frag, terminals].values())
		root = frag[1:frag.index(' ')]
		nonterms = frag.count('(') - 1
		# Sangati & Zuidema (2011, eq. 5)
		# FIXME: verify that this formula is equivalent to Bod (2003).
		ewe = sum(Fraction(v, fragmentcount[k])
				for k, v in fragments[frag, terminals].items())
		# Bonnema (2003, p. 34)
		bon = 2 ** -nonterms * (freq / ntfd[root])
		short = 0.5
		return freq, ewe, bon, short
	uniformweight = (1, 1, 1, 1)
	grammar = {}
	backtransform = {}
	ids = UniqueIDs()
	# build index of the number of fragments extracted from a tree for ewe
	fragmentcount = defaultdict(int)
	for indices in fragments.values():
		for index, cnt in indices.items():
			fragmentcount[index] += cnt
	# ntfd: frequency of a non-terminal node in treebank
	ntfd = multiset(node.label for tree in trees for node in tree.subtrees())

	# binarize, turn to lcfrs productions
	# use artificial markers of binarization as disambiguation,
	# construct a mapping of productions to fragments
	for frag, terminals in fragments:
		prods, newfrag = flatten(frag, terminals, ids)
		prod = prods[0]
		if prod[0][1] == 'Epsilon':  # lexical production
			grammar[prod] = getweight(frag, terminals)
			continue
		elif prod in backtransform:
			# normally, rules of fragments are disambiguated by binarization IDs
			# in case there's a fragment with only one or two frontier nodes,
			# we add an artficial node.
			newlabel = "%s}<%d>%s" % (prod[0][0], next(ids),
					'' if len(prod[1]) == 1 else '_%d' % len(prod[1]))
			prod1 = ((prod[0][0], newlabel) + prod[0][2:], prod[1])
			# we have to determine fanout of the first nonterminal
			# on the right hand side
			prod2 = ((newlabel, prod[0][1]),
				tuple((0,) for component in prod[1]
				for a in component if a == 0))
			prods[:1] = [prod1, prod2]

		# first binarized production gets prob. mass
		grammar[prod] = getweight(frag, terminals)
		grammar.update(zip(prods[1:], repeat(uniformweight)))
		# & becomes key in backtransform
		backtransform[prod] = newfrag
	if debug:
		ids = count()
		flatfrags = [flatten(frag, terminals, ids)
				for frag, terminals in fragments]
		print("recurring fragments:")
		for a, b in zip(flatfrags, fragments):
			print("fragment: %s\nprod:     %s" % (b[0], "\n\t".join(
				printrule(r, yf, 0) for r, yf in a[0])))
			print("template: %s\nfreq: %2d  sent: %s\n" % (
					a[1], len(fragments[b]), ' '.join('_' if x is None
					else quotelabel(x) for x in b[1])))
		print("backtransform:")
		for a, b in backtransform.items():
			print(a, b)

	# order grammar such that we have these clusters:
	# 1. non-binarized rules or initial rules of a binarized constituent
	# 2: non-initial binarized rules.
	# 3: lexical productions sorted by word
	# this is so that the backtransform aligns with the first part of the rules
	grammar = sorted(grammar.items(), key=lambda rule: (
				rule[0][1][0] if rule[0][0][1] == 'Epsilon' else '',  # word?
				'}<' in rule[0][0][0],  # fragment-internal production?
				rule[0][0]))  # LHS
	# replace keys with numeric ids of rules, drop terminals.
	backtransform = [backtransform[r] for r, _ in grammar
			if r in backtransform]
	# relative frequences as probabilities (don't normalize shortest & bon)
	ntsums = defaultdict(int)
	ntsumsewe = defaultdict(int)
	for rule, (freq, ewe, _, _) in grammar:
		ntsums[rule[0][0]] += freq
		ntsumsewe[rule[0][0]] += ewe
	eweweights = [float(ewe) / ntsumsewe[rule[0][0]]
			for rule, (_, ewe, _, _) in grammar]
	bonweights = [bon for rule, (_, _, bon, _) in grammar]
	shortest = [s for rule, (_, _, _, s) in grammar]
	grammar = [(rule, Fraction(freq, ntsums[rule[0][0]]))
			for rule, (freq, _, _, _) in grammar]
	return grammar, backtransform, dict(
			ewe=eweweights, bon=bonweights, shortest=shortest)


def sortgrammar(grammar):
	""" put lexical rules in the end and sort by word. """
	return sorted(grammar, key=lambda rule:
			(rule[0][1][0], rule[0][0])  # word, LHS
			if rule[0][0][1] == 'Epsilon'
			else ('', rule[0][0]))  # phrasal rule: LHS


def coarse_grammar(trees, sents, level=0):
	""" collapse all labels to X except ROOT and POS tags. """
	if level == 0:
		repl = lambda x: "X"
	label = re.compile("[^^|<>-]+")
	for tree in trees:
		for subtree in tree.subtrees():
			if subtree.label != "ROOT" and isinstance(subtree[0], Tree):
				subtree.label = label.sub(repl, subtree.label)
	return treebankgrammar(trees, sents)


def nodefreq(tree, dectree, subtreefd, nonterminalfd):
	""" Auxiliary function for DOP reduction.
	Counts frequencies of nodes and calculate the number of
	subtrees headed by each node. updates "subtreefd" and "nonterminalfd"
	as a side effect. Expects a normal tree and a tree with IDs.

	:param subtreefd: the multiset to store the counts of subtrees
	:param nonterminalfd: the multiset to store the counts of non-terminals

	>>> fd = multiset()
	>>> d = TreeDecorator()
	>>> tree = Tree("(S (NP mary) (VP walks))")
	>>> dectree = d.decorate(tree, ['mary', 'walks'])
	>>> nodefreq(tree, dectree, fd, multiset())
	4
	>>> fd == multiset({'S': 4, 'NP': 1, 'VP': 1, 'NP@1-0': 1, 'VP@1-1': 1})
	True """
	nonterminalfd[tree.label] += 1
	nonterminalfd[dectree.label] += 1
	if isinstance(tree[0], Tree):
		n = reduce(mul, (nodefreq(x, ux, subtreefd, nonterminalfd) + 1
			for x, ux in zip(tree, dectree)))
	else:  # lexical production
		n = 1
	subtreefd[tree.label] += n
	# only add counts when dectree.label is actually an interior node,
	# e.g., root node receives no ID so shouldn't be counted twice
	if dectree.label != tree.label:  # if subtreefd[dectree.label] == 0:
		# NB: assignment, not addition; addressed nodes should be unique, and
		# with packed graph encoding we don't want duplicate counts.
		subtreefd[dectree.label] = n
	return n


class TreeDecorator(object):
	""" Auxiliary class for DOP reduction.
	Adds unique identifiers to each internal non-terminal of a tree.
	If initialized with ``memoize=True``, equivalent subtrees will get the
	same identifiers. """
	def __init__(self, memoize=False):
		self.ids = self.n = 0
		self.packedgraphs = {}
		self.memoize = memoize

	def decorate(self, tree, sent):
		""" Return a copy of tree with labels decorated with IDs.

		>>> d = TreeDecorator()
		>>> tree = Tree.parse("(S (NP (DT 0) (N 1)) (VP 2))", parse_leaf=int)
		>>> d.decorate(tree, ['the', 'dog', 'walks'])
		Tree('S', [Tree('NP@1-0', [Tree('DT@1-1', [0]),
				Tree('N@1-2', [1])]), Tree('VP@1-3', [2])])
		>>> d = TreeDecorator(memoize=True)
		>>> print(d.decorate(Tree.parse("(S (NP (DT 0) (N 1)) (VP 2))",
		...		parse_leaf=int), ['the', 'dog', 'walks']))
		(S (NP@1-1 (DT@1-2 0) (N@1-3 1)) (VP@1-4 2))
		>>> print(d.decorate(Tree.parse("(S (NP (DT 0) (N 1)) (VP 2))",
		...		parse_leaf=int), ['the', 'dog', 'barks']))
		(S (NP@1-1 (DT@1-2 0) (N@1-3 1)) (VP@2-4 2)) """
		self.n += 1  # sentence number
		if self.memoize:
			self.ids = 0  # node number
			# wrap tree to get equality wrt sent
			tree = DiscTree(tree.freeze(), sent)
			dectree = ImmutableTree(tree.label, map(self._recdecorate, tree))
		else:
			dectree = Tree.convert(tree.copy(True))
			#skip top node, should not get an ID
			for m, a in enumerate(islice(dectree.subtrees(), 1, None)):
				a.label = "%s@%d-%d" % (a.label, self.n, m)
		return dectree

	def _recdecorate(self, tree):
		""" Traverse subtrees not yet seen. """
		if isinstance(tree, int):
			return tree
		elif tree not in self.packedgraphs:
			self.ids += 1
			self.packedgraphs[tree] = ImmutableTree(("%s@%d-%d" % (
					tree.label, self.n, self.ids)),
					[self._recdecorate(child) for child in tree])
			return self.packedgraphs[tree]
		return self._copyexceptindices(tree, self.packedgraphs[tree])

	def _copyexceptindices(self, tree1, tree2):
		""" Copy the nonterminals from tree2, but take indices from tree1. """
		if not isinstance(tree1, Tree):
			return tree1
		self.ids += 1
		return ImmutableTree(tree2.label,
			[self._copyexceptindices(a, b) for a, b in zip(tree1, tree2)])


class DiscTree(ImmutableTree):
	""" Wrap an immutable tree with indices as leaves
	and a sentence. Provides hash & equality.  """
	def __init__(self, tree, sent):
		self.sent = tuple(sent)
		super(DiscTree, self).__init__(tree.label, tuple(
				DiscTree(child, sent) if isinstance(child, Tree) else child
				for child in tree))

	def __eq__(self, other):
		return isinstance(other, Tree) and eqtree(self, self.sent,
				other, other.sent)

	def __hash__(self):
		return hash((self.label, ) + tuple(child.__hash__()
				if isinstance(child, Tree) else self.sent[child]
				for child in self))

	def __repr__(self):
		return "DisctTree(%r, %r)" % (
				super(DiscTree, self).__repr__(), self.sent)


def eqtree(tree1, sent1, tree2, sent2):
	""" Test whether two discontinuous trees are equivalent;
	assumes canonicalized() ordering. """
	if tree1.label != tree2.label or len(tree1) != len(tree2):
		return False
	for a, b in zip(tree1, tree2):
		istree = isinstance(a, Tree)
		if istree != isinstance(b, Tree):
			return False
		elif not istree:
			return sent1[a] == sent2[b]
		elif not a.__eq__(b):
			return False
	return True


def quotelabel(label):
	""" Escapes two things: parentheses and non-ascii characters.
	Parentheses are replaced by square brackets. Also escapes non-ascii
	characters, so that phrasal labels can remain ascii-only. """
	newlabel = label.replace('(', '[').replace(')', ']')
	# juggling to get str in both Python 2 and Python 3.
	return str(newlabel.encode('unicode-escape').decode('ascii'))

FRONTIERORTERM_new = re.compile(r"\([^ ]+(?: [0-9]+)+\)")


def new_flatten(tree, sent, ids):
	""" Auxiliary function for Double-DOP.
	Remove internal nodes from a tree and read off its binarized
	productions. Aside from returning productions, also return tree with
	lexical and frontier nodes replaced by a templating symbol '%s'.
	Input is a tree and sentence, as well as an iterator which yields
	unique IDs for non-terminals introdudced by the binarization;
	output is a tuple (prods, frag). Trees are in the form of strings.

	#>>> ids = count()
	#>>> sent = [None, ',', None, '.']
	#>>> tree = "(ROOT (S_2 0 2) (ROOT|<$,>_2 ($, 1) ($. 3)))"
	#>>> new_flatten(tree, sent, ids)
	#([(('ROOT', 'ROOT}<0>', '$.@.'), ((0, 1),)),
	#(('ROOT}<0>', 'S_2', '$,@,'), ((0, 1, 0),)),
	#(('$,@,', 'Epsilon'), (',',)), (('$.@.', 'Epsilon'), ('.',))],
	#'(S_2 {0}) (ROOT|<$,>_2 ($, {1}) ($. {2}))',
	#['(S_2 ', 0, ') (ROOT|<$,>_2 ($, ', 1, ') ($. ', 2 '))']) """
	from discodop.treetransforms import factorconstituent, addbitsets

	def repl(x):
		""" Add information to a frontier or terminal:

		:frontiers: ``(label indices)``
		:terminals: ``(tag@word idx)`` """
		n = x.group(2)  # index w/leading space
		nn = int(n)
		if sent[nn] is None:
			return x.group(0)  # (label indices)
		word = quotelabel(sent[nn])
		# (tag@word idx)
		return "(%s@%s%s)" % (x.group(1), word, n)

	if tree.count(' ') == 1:
		return lcfrs_productions(addbitsets(tree), sent), ([str(tree)], [])
	# give terminals unique POS tags
	prod = FRONTIERORTERM.sub(repl, tree)
	# remove internal nodes, reorder
	prod = "%s %s)" % (prod[:prod.index(' ')],
		' '.join(x.group(0) for x in sorted(FRONTIERORTERM.finditer(prod),
		key=lambda x: int(x.group(2)))))
	prods = lcfrs_productions(factorconstituent(addbitsets(prod),
			"}", factor='left', markfanout=True, ids=ids, threshold=2), sent)

	# remember original order of frontiers / terminals for template
	order = [int(x.group(2)) for x in FRONTIERORTERM.finditer(prod)]
	# ensure string, split around substitution sites.
	#lambda x: order[x.group(2)],
	treeparts = FRONTIERORTERM_new.split(str(tree))
	return prods, (treeparts, order)


class UniqueIDs(object):
	""" Produce numeric IDs. Can be used as iterator (ID will not be re-used)
	and dictionary (ID will be re-used for same key).

	>>> ids = UniqueIDs()
	>>> next(ids)
	0
	>>> ids['foo'], ids['bar'], ids['foo']
	(1, 2, 1) """
	def __init__(self):
		self.cnt = 0  # next available ID
		self.ids = {}  # IDs for labels seen

	def __getitem__(self, key):
		val = self.ids.get(key)
		if val is None:
			val = self.ids[key] = self.cnt
			self.cnt += 1
		return val

	def __next__(self):
		self.cnt += 1
		return self.cnt - 1

	def __iter__(self):
		return self

	next = __next__

FRONTIERORTERM = re.compile(r"\(([^ ]+)( [0-9]+)(?: [0-9]+)*\)")


def flatten(tree, sent, ids):
	""" Auxiliary function for Double-DOP.
	Remove internal nodes from a tree and read off the binarized
	productions of the resulting flattened tree. Aside from returning
	productions, also return tree with lexical and frontier nodes replaced by a
	templating symbol '{n}' where n is an index.
	Input is a tree and sentence, as well as an iterator which yields
	unique IDs for non-terminals introdudced by the binarization;
	output is a tuple (prods, frag). Trees are in the form of strings.

	>>> ids = UniqueIDs()
	>>> sent = [None, ',', None, '.']
	>>> tree = "(ROOT (S_2 0 2) (ROOT|<$,>_2 ($, 1) ($. 3)))"
	>>> flatten(tree, sent, ids)
	([(('ROOT', 'ROOT}<0>', '$.@.'), ((0, 1),)),
	(('ROOT}<0>', 'S_2', '$,@,'), ((0, 1, 0),)),
	(('$,@,', 'Epsilon'), (',',)), (('$.@.', 'Epsilon'), ('.',))],
	'(ROOT {0} (ROOT|<$,>_2 {1} {2}))')
	>>> flatten("(NN 0)", ["foo"], ids)
	([(('NN', 'Epsilon'), ('foo',))], '(NN 0)')
	>>> flatten(r"(S (S|<VP> (S|<NP> (NP (ART 0) (CNP (CNP|<TRUNC> "
	... "(TRUNC 1) (CNP|<KON> (KON 2) (CNP|<NN> (NN 3)))))) (S|<VAFIN> "
	... "(VAFIN 4))) (VP (VP|<ADV> (ADV 5) (VP|<NP> (NP (ART 6) (NN 7)) "
	... "(VP|<NP> (NP_2 8 10) (VP|<VVPP> (VVPP 9))))))))",
	... ['Das', 'Garten-', 'und', 'Friedhofsamt', 'hatte', 'kuerzlich',
	... 'dem', 'Ortsbeirat', None, None, None], ids)
	([(('S', 'S}<8>_2', 'VVPP'), ((0, 1, 0),)),
	(('S}<8>_2', 'S}<7>', 'NP_2'), ((0, 1), (1,))),
	(('S}<7>', 'S}<6>', 'NN@Ortsbeirat'), ((0, 1),)),
	(('S}<6>', 'S}<5>', 'ART@dem'), ((0, 1),)),
	(('S}<5>', 'S}<4>', 'ADV@kuerzlich'), ((0, 1),)),
	(('S}<4>', 'S}<3>', 'VAFIN@hatte'), ((0, 1),)),
	(('S}<3>', 'S}<2>', 'NN@Friedhofsamt'), ((0, 1),)),
	(('S}<2>', 'S}<1>', 'KON@und'), ((0, 1),)),
	(('S}<1>', 'ART@Das', 'TRUNC@Garten-'), ((0, 1),)),
	(('ART@Das', 'Epsilon'), ('Das',)),
	(('TRUNC@Garten-', 'Epsilon'), ('Garten-',)),
	(('KON@und', 'Epsilon'), ('und',)),
	(('NN@Friedhofsamt', 'Epsilon'), ('Friedhofsamt',)),
	(('VAFIN@hatte', 'Epsilon'), ('hatte',)),
	(('ADV@kuerzlich', 'Epsilon'), ('kuerzlich',)),
	(('ART@dem', 'Epsilon'), ('dem',)),
	(('NN@Ortsbeirat', 'Epsilon'), ('Ortsbeirat',))],
	'(S (S|<VP> (S|<NP> (NP {0} (CNP (CNP|<TRUNC> {1} (CNP|<KON> {2} \
	(CNP|<NN> {3}))))) (S|<VAFIN> {4})) (VP (VP|<ADV> {5} (VP|<NP> \
	(NP {6} {7}) (VP|<NP> {8} (VP|<VVPP> {9})))))))')
	>>> flatten("(S|<VP>_2 (VP_3 (VP|<NP>_3 (NP 0) (VP|<ADV>_2 "
	... "(ADV 2) (VP|<VVPP> (VVPP 4))))) (S|<VAFIN> (VAFIN 1)))",
	... (None, None, None, None, None), ids)
	([(('S|<VP>_2', 'S|<VP>_2}<10>', 'VVPP'), ((0,), (1,))),
	(('S|<VP>_2}<10>', 'S|<VP>_2}<9>', 'ADV'), ((0, 1),)),
	(('S|<VP>_2}<9>', 'NP', 'VAFIN'), ((0, 1),))],
	'(S|<VP>_2 (VP_3 (VP|<NP>_3 {0} (VP|<ADV>_2 {2} (VP|<VVPP> {3})))) \
	(S|<VAFIN> {1}))') """
	from discodop.treetransforms import factorconstituent, addbitsets

	def repl(x):
		""" Add information to a frontier or terminal:

		:frontiers: ``(label indices)``
		:terminals: ``(tag@word idx)`` """
		n = x.group(2)  # index w/leading space
		nn = int(n)
		if sent[nn] is None:
			return x.group(0)  # (label indices)
		word = quotelabel(sent[nn])
		# (tag@word idx)
		return "(%s@%s%s)" % (x.group(1), word, n)

	if tree.count(' ') == 1:
		return lcfrs_productions(addbitsets(tree), sent), str(tree)
	# give terminals unique POS tags
	prod = FRONTIERORTERM.sub(repl, tree)
	# remove internal nodes, reorder
	prod = "%s %s)" % (prod[:prod.index(' ')],
			' '.join(x.group(0) for x in sorted(FRONTIERORTERM.finditer(prod),
			key=lambda x: int(x.group(2)))))
	prods = lcfrs_productions(factorconstituent(addbitsets(prod), "}",
			factor='left', markfanout=True, markyf=True, ids=ids, threshold=2),
			sent)
	# remember original order of frontiers / terminals for template
	order = {x.group(2): "{%d}" % n
			for n, x in enumerate(FRONTIERORTERM.finditer(prod))}
	# mark substitution sites and ensure string.
	newtree = FRONTIERORTERM.sub(lambda x: order[x.group(2)], tree)
	return prods, str(newtree)


def rangeheads(s):
	""" Iterate over a sequence of numbers and return first element of each
	contiguous range. Input should be shorted.

	>>> rangeheads( (0, 1, 3, 4, 6) )
	[0, 3, 6]
	"""
	sset = set(s)
	return [a for a in s if a - 1 not in sset]


def ranges(s):
	""" Partition s into a sequence of lists corresponding to contiguous ranges

	>>> list(ranges( (0, 1, 3, 4, 6) ))
	[[0, 1], [3, 4], [6]] """
	rng = []
	for a in s:
		if not rng or a == rng[-1] + 1:
			rng.append(a)
		else:
			yield rng
			rng = [a]
	if rng:
		yield rng


def defaultparse(wordstags, rightbranching=False):
	""" a default parse, either right branching NPs, or all words under a single
	constituent 'NOPARSE'.

	>>> defaultparse([('like','X'), ('this','X'), ('example', 'NN'),
	... ('here','X')])
	'(NOPARSE (X like) (X this) (NN example) (X here))'
	>>> defaultparse([('like','X'), ('this','X'), ('example', 'NN'),
	... ('here','X')], True)
	'(NP (X like) (NP (X this) (NP (NN example) (NP (X here)))))' """
	if rightbranching:
		if wordstags[1:]:
			return "(NP (%s %s) %s)" % (wordstags[0][1],
					wordstags[0][0], defaultparse(wordstags[1:], rightbranching))
		return "(NP (%s %s))" % wordstags[0][::-1]
	return "(NOPARSE %s)" % ' '.join("(%s %s)" % a[::-1] for a in wordstags)


def printrule(r, yf, w):
	""" :returns: a string with a representation of a rule. """
	return "%s %s --> %s\t %r" % (w, r[0], ' '.join(x for x in r[1:]), list(yf))


def cartpi(seq):
	""" itertools.product doesn't support infinite sequences!

	>>> list(islice(cartpi([count(), count(0)]), 9))
	[(0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6), (0, 7), (0, 8)] """
	if seq:
		return (b + (a, ) for b in cartpi(seq[:-1]) for a in seq[-1])
	return ((), )


def write_lncky_grammar(rules, lexicon, out, encoding='utf-8'):
	""" Takes a bitpar grammar and converts it to the format of
	Mark Jonhson's cky parser. """
	grammar = []
	for a in io.open(rules, encoding=encoding):
		a = a.split()
		p, rule = a[0], a[1:]
		grammar.append('%s %s --> %s\n' % (p, rule[0], ' '.join(rule[1:])))
	for a in io.open(lexicon, encoding=encoding):
		a = a.split()
		word, tags = a[0], a[1:]
		tags = zip(tags[::2], tags[1::2])
		grammar.extend('%s %s --> %s\n' % (p, t, word) for t, p in tags)
	assert 'VROOT' in grammar[0]
	io.open(out, 'w', encoding=encoding).writelines(grammar)


def write_lcfrs_grammar(grammar, bitpar=False):
	""" Writes a grammar in a simple text file format. Rules are written in
	the order as they appear in the sequence `grammar`, except that the lexicon
	file lists words in sorted order (with tags for each word in the order of
	`grammar`). For a description of the file format, see ``grammar.FORMAT``.

	:param grammar:  a sequence of rule tuples, as produced by
		treebankgrammar(), dopreduction(), or doubledop().
	:param bitpar: when ``True``, use bitpar format: for rules, put weight
		first (as decimal fraction or frequency) and leave out the yield
		function.
	:returns: rules, lexicon; bytes object & a unicode string, respectively """
	# when grammar is a PLCFRS, write rational fractions.
	# when grammar is bitpar PCFG, write frequencies if probabilities sum to 1,
	# i.e., in that case probalities can be re-computed as relative
	# frequencies. otherwise, resort to decimal floats (imprecise).
	rules, lexicon = [], []
	lexical = {}
	if bitpar:
		ntfd = defaultdict(int)
		for (r, _), w in grammar:
			ntfd[r[0]] += w
	freqs = bitpar and all(a == 1 for a in ntfd.values())
	for (r, yf), w in grammar:
		if len(r) == 2 and r[1] == 'Epsilon':
			lexical.setdefault(unicode(yf[0]), []).append((r[0], w))
			continue
		elif bitpar:
			rules.append(('%g\t%s\n' % (w.numerator if freqs else w,
					'\t'.join(x for x in r))))
		else:
			yfstr = ','.join(''.join(map(str, a)) for a in yf)
			rules.append(('%s\t%s\t%s\n' % (
					'\t'.join(x for x in r), yfstr,
					w.numerator if freqs else w)))
	for word in sorted(lexical):
		lexicon.append(word)
		for tag, w in lexical[word]:
			if freqs:
				lexicon.append(unicode('\t%s %d' % (tag, w.numerator)))
			else:
				lexicon.append(unicode('\t%s %s' % (tag,
						(float(w) if bitpar else w))))
		lexicon.append(unicode('\n'))
	return ''.join(rules).encode('ascii'), u''.join(lexicon)


def subsetgrammar(a, b):
	""" test whether grammar a is a subset of b. """
	difference = set(map(itemgetter(0), a)) - set(map(itemgetter(0), b))
	if not difference:
		return True
	print("missing productions:")
	for r, yf in difference:
		print(printrule(r, yf, 0.0))
	return False


def grammarinfo(grammar, dump=None):
	""" print(some statistics on a grammar, before it goes through Grammar().)

	:param dump: if given a filename, will dump distribution of parsing
		complexity to a file (i.e., p.c. 3 occurs 234 times, 4 occurs 120
		times, etc.) """
	from discodop.eval import mean
	lhs = {rule[0] for (rule, yf), w in grammar}
	l = len(grammar)
	result = "labels: %d" % len({rule[a] for (rule, yf), w in grammar
							for a in range(3) if len(rule) > a})
	result += " of which preterminals: %d\n" % (
		len({rule[0] for (rule, yf), w in grammar if rule[1] == 'Epsilon'})
		or len({rule[a] for (rule, yf), w in grammar
				for a in range(1, 3) if len(rule) > a and rule[a] not in lhs}))
	ll = sum(1 for (rule, yf), w in grammar if rule[1] == 'Epsilon')
	result += "clauses: %d  lexical clauses: %d" % (l, ll)
	result += " non-lexical clauses: %d\n" % (l - ll)
	n, r, yf, w = max((len(yf), rule, yf, w) for (rule, yf), w in grammar)
	result += "max fan-out: %d in " % n
	result += printrule(r, yf, w)
	result += " average: %g\n" % mean([len(yf) for (_, yf), _, in grammar])
	n, r, yf, w = max((sum(map(len, yf)), rule, yf, w)
				for (rule, yf), w in grammar if rule[1] != 'Epsilon')
	result += "max variables: %d in %s\n" % (n, printrule(r, yf, w))

	def parsingcomplexity(yf):
		""" this sums the fanouts of LHS & RHS """
		if isinstance(yf[0], tuple):
			return len(yf) + sum(map(len, yf))
		return 1  # NB: a lexical production has complexity 1

	pc = {(rule, yf, w): parsingcomplexity(yf)
							for (rule, yf), w in grammar}
	r, yf, w = max(pc, key=pc.get)
	result += "max parsing complexity: %d in %s" % (
			pc[r, yf, w], printrule(r, yf, w))
	result += " average %g" % mean(pc.values())
	if dump:
		pcdist = multiset(pc.values())
		open(dump, "w").writelines("%d\t%d\n" % x for x in pcdist.items())
	return result


def test():
	""" Run some tests. """
	from discodop import plcfrs
	from discodop._grammar import Grammar
	from discodop.treebank import NegraCorpusReader
	from discodop.treetransforms import binarize, unbinarize, \
			addfanoutmarkers, removefanoutmarkers
	from discodop.disambiguation import recoverfragments
	from discodop.kbest import lazykbest
	from discodop.agenda import getkey
	from discodop.fragments import getfragments
	logging.basicConfig(level=logging.DEBUG, format='%(message)s')
	filename = "alpinosample.export"
	corpus = NegraCorpusReader('.', filename, punct='move')
	sents = list(corpus.sents().values())
	trees = [addfanoutmarkers(binarize(a.copy(True), horzmarkov=1))
			for a in list(corpus.parsed_sents().values())[:10]]

	print('plcfrs')
	lcfrs = Grammar(treebankgrammar(trees, sents), start=trees[0].label)
	print(lcfrs)

	print('dop reduction')
	grammar = Grammar(dopreduction(trees[:2], sents[:2])[0],
			start=trees[0].label)
	print(grammar)
	grammar.testgrammar()

	fragments = getfragments(trees, sents, 1)
	debug = '--debug' in sys.argv
	grammarx, backtransform, _ = doubledop(trees, fragments, debug=debug)
	print('\ndouble dop grammar')
	grammar = Grammar(grammarx, start=trees[0].label)
	grammar.getmapping(grammar, striplabelre=None,
			neverblockre=re.compile(b'^#[0-9]+|.+}<'),
			splitprune=False, markorigin=False)
	print(grammar)
	assert grammar.testgrammar(), "DOP1 should sum to 1."
	for tree, sent in zip(corpus.parsed_sents().values(), sents):
		print("sentence:", ' '.join(a.encode('unicode-escape').decode()
				for a in sent))
		chart, start, msg = plcfrs.parse(sent, grammar, exhaustive=True)
		print('\n', msg, end='')
		print("\ngold ", tree)
		print("double dop", end='')
		if start:
			mpp = {}
			parsetrees = {}
			derivations, D, _ = lazykbest(chart, start, 1000,
				grammar.tolabel, b'}<')
			for d, (t, p) in zip(D[start], derivations):
				r = Tree(recoverfragments(getkey(d), D,
					grammar, backtransform))
				r = str(removefanoutmarkers(unbinarize(r)))
				mpp[r] = mpp.get(r, 0.0) + exp(-p)
				parsetrees.setdefault(r, []).append((t, p))
			print(len(mpp), 'parsetrees', end='')
			print(sum(map(len, parsetrees.values())), 'derivations')
			for t, tp in sorted(mpp.items(), key=itemgetter(1)):
				print(tp, '\n', t, end='')
				print("match:", t == str(tree))
				assert len(set(parsetrees[t])) == len(parsetrees[t])
				if not debug:
					continue
				for deriv, p in sorted(parsetrees[t], key=itemgetter(1)):
					print(' <= %6g %s' % (exp(-p), deriv))
		else:
			print("no parse")
			plcfrs.pprint_chart(chart, sent, grammar.tolabel)
		print()
	tree = Tree.parse("(ROOT (S (F (E (S (C (B (A 0))))))))", parse_leaf=int)
	Grammar(treebankgrammar([tree], [[str(a) for a in range(10)]]))


def main():
	""" Command line interface to create grammars from treebanks. """
	import gzip
	from getopt import gnu_getopt, GetoptError
	from discodop.treetransforms import addfanoutmarkers, canonicalize
	from discodop.treebank import getreader, splitpath
	from discodop.fragments import getfragments
	logging.basicConfig(level=logging.DEBUG, format='%(message)s')
	shortoptions = ''
	flags = ('gzip', 'packed')
	options = ('inputfmt=', 'inputenc=', 'dopestimator=', 'numproc=')
	try:
		opts, args = gnu_getopt(sys.argv[1:], shortoptions, flags + options)
		model, treebankfile, grammarfile = args
	except (GetoptError, ValueError) as err:
		print('error: %r\n%s' % (err, USAGE))
		sys.exit(2)
	opts = dict(opts)
	assert model in ('pcfg', 'plcfrs', 'dopreduction', 'doubledop'), (
		'unrecognized model: %r' % model)
	assert opts.get('dopestimator', 'dop1') in ('dop1', 'ewe', 'shortest'), (
		'unrecognized estimator: %r' % opts['dopestimator'])

	# read treebank
	reader = getreader(opts.get('--inputfmt', 'export'))
	corpus = reader(*splitpath(treebankfile),
			encoding=opts.get('--inputenc', 'utf8'))
	trees = list(corpus.parsed_sents().values())
	sents = list(corpus.sents().values())
	for a in trees:
		canonicalize(a)
		addfanoutmarkers(a)

	# read off grammar
	if model in ('pcfg', 'plcfrs'):
		grammar = treebankgrammar(trees, sents)
	elif model == 'dopreduction':
		grammar, altweights = dopreduction(trees, sents,
				packedgraph='--packed' in opts)
	elif model == 'doubledop':
		numproc = int(opts.get('--numproc', 1))
		fragments = getfragments(trees, sents, numproc)
		grammar, backtransform, altweights = doubledop(trees, fragments)
	if opts.get('--dopestimator', 'dop1') == 'ewe':
		grammar = [(rule, w) for (rule, _), w in
				zip(grammar, altweights['ewe'])]
	elif opts.get('--dopestimator', 'dop1') == 'shortest':
		grammar = [(rule, w) for (rule, _), w in
				zip(grammar, altweights['shortest'])]

	print(grammarinfo(grammar))
	rules = grammarfile + '.rules'
	lexicon = grammarfile + '.lex'
	if '--gzip' in opts:
		myopen = gzip.open
		rules += '.gz'
		lexicon += '.gz'
	else:
		myopen = open
	bitpar = model == 'pcfg' or opts.get('--inputfmt') == 'bracket'
	rules, lexicon = write_lcfrs_grammar(grammar, bitpar=bitpar)
	try:
		from discodop._grammar import Grammar
	except ImportError:
		pass
	else:
		cgrammar = Grammar(rules, lexicon)
		cgrammar.testgrammar()
	# write output
	with myopen(rules, 'w') as rulesfile:
		rulesfile.write(rules)
	with codecs.getwriter('utf-8')(myopen(lexicon, 'w')) as lexiconfile:
		lexiconfile.write(lexicon)
	if model == 'doubledop':
		backtransformfile = '%s.backtransform%s' % (grammarfile,
			'.gz' if '--gzip' in opts else '')
		myopen(backtransformfile, 'w').writelines(
				'%s\n' % a for a in backtransform)
		print('wrote backtransform to', backtransformfile)
	print('wrote grammar to %s and %s.' % (rules, lexicon))

if __name__ == '__main__':
	if '--test' in sys.argv:
		test()
	else:
		main()



""" Implementation of Huang & Chiang (2005): Better k-best parsing. """
from __future__ import print_function
from discodop.agenda import Agenda
from discodop.containers import ChartItem, Edge, RankedEdge

cimport cython
from discodop.agenda cimport Entry, Agenda, nsmallest
from discodop.containers cimport ChartItem, SmallChartItem, CFGChartItem, \
		Edge, Rule, CFGEdge, LCFRSEdge, RankedEdge, RankedCFGEdge, \
		new_LCFRSEdge, UInt, UChar

cdef tuple unarybest = (0, ), binarybest = (0, 0)

cdef inline getcandidates(dict chart, ChartItem v, int k):
	""" :returns: a heap with up to k candidate arcs starting from vertex v """
	# NB: the priority queue should either do a stable sort, or should
	# sort on rank vector as well to have ties resolved in FIFO order;
	# otherwise the sequence (0, 0) -> (1, 0) -> (1, 1) -> (0, 1) -> (1, 1)
	# can occur (given that the first two have probability x and the latter
	# three probability y), in which case insertion order should count.
	# Otherwise (1, 1) ends up in D[v] after which (0. 1) generates it
	# as a neighbor and puts it in cand[v] for a second time.
	cdef LCFRSEdge el
	if v not in chart:
		return Agenda()  # raise error?
	return Agenda([
		(RankedEdge(v, el, 0, 0 if el.right.label else -1), el.inside)
				for el in nsmallest(k, chart[v].values())])


@cython.wraparound(True)
cpdef inline lazykthbest(ChartItem v, int k, int k1, dict D, dict cand,
		dict chart, set explored):
	cdef Entry entry
	cdef RankedEdge ej
	# k1 is the global k
	# first visit of vertex v?
	if v not in cand:
		# initialize the heap
		cand[v] = getcandidates(chart, v, k1)
	while v not in D or len(D[v]) < k:
		if v in D:
			# last derivation
			entry = D[v][-1]
			ej = entry.key
			# update the heap, adding the successors of last derivation
			lazynext(ej, k1, D, cand, chart, explored)
		# get the next best derivation and delete it from the heap
		if cand[v]:
			D.setdefault(v, []).append((<Agenda>cand[v]).popentry())
		else:
			break
	return D


cdef inline lazynext(RankedEdge ej, int k1, dict D, dict cand, dict chart,
		set explored):
	cdef RankedEdge ej1
	cdef double prob
	# add the |e| neighbors
	for i in range(2):
		if i == 0:
			ei = ej.edge.left
			ej1 = RankedEdge(ej.head, ej.edge, ej.left + 1, ej.right)
		elif i == 1 and ej.right >= 0:  # edge.right.label:
			ei = ej.edge.right
			ej1 = RankedEdge(ej.head, ej.edge, ej.left, ej.right + 1)
		else:
			break
		# recursively solve a subproblem
		# NB: increment j1[i] again because j is zero-based and k is not
		lazykthbest(ei, (ej1.right if i else ej1.left) + 1, k1,
							D, cand, chart, explored)
		# if it exists and is not in heap yet
		if ((ei in D and (ej1.right if i else ej1.left) < len(D[ei]))
			and ej1 not in explored):  # cand[ej1.head]): <= gives duplicates
			prob = getprob(chart, D, ej1)
			# add it to the heap
			cand[ej1.head][ej1] = prob
			explored.add(ej1)


cdef inline double getprob(dict chart, dict D, RankedEdge ej) except -1.0:
	""" Get inside probability of 'ej'; try looking in D, or else use chart. """
	cdef ChartItem ei
	cdef Edge edge
	cdef Entry entry
	cdef double result, prob
	ei = ej.edge.left
	if ei in D:
		entry = D[ei][ej.left]
		prob = entry.value
	elif ej.left == 0:
		edge = min(chart[ei])
		prob = edge.inside
	else:
		raise ValueError("non-zero rank vector not part of explored derivations")
	result = ej.edge.rule.prob + prob
	if ej.right >= 0:  # if e.right.label:
		ei = ej.edge.right
		if ei in D:
			entry = D[ei][ej.right]
			prob = entry.value
		elif ej.right == 0:
			edge = min(chart[ei])
			prob = edge.inside
		else:
			raise ValueError(
			"non-zero rank vector not part of explored derivations")
		result += prob
	return result


# --- start CFG specific
cdef inline getcandidatescfg(list chart, UInt label,
		UChar start, UChar end, int k):
	""" :returns: a heap with up to k candidate arcs starting from vertex v """
	# NB: the priority queue should either do a stable sort, or should
	# sort on rank vector as well to have ties resolved in FIFO order;
	# otherwise the sequence (0, 0) -> (1, 0) -> (1, 1) -> (0, 1) -> (1, 1)
	# can occur (given that the first two have probability x and the latter
	# three probability y), in which case insertion order should count.
	# Otherwise (1, 1) ends up in D[v] after which (0. 1) generates it
	# as a neighbor and puts it in cand[v] for a second time.
	cdef CFGEdge ec
	cell = chart[start][end]
	if not cell.get(label):
		return Agenda()
	return Agenda(
		[(RankedCFGEdge(label, start, end, ec, 0, 0 if ec.rule is not NULL
						and ec.rule.rhs2 else -1), ec.inside)
						for ec in nsmallest(k, cell[label].values())])


@cython.wraparound(True)
cpdef inline lazykthbestcfg(UInt label, UChar start, UChar end, int k, int k1,
		list D, list cand, list chart, set explored):
	cdef Entry entry
	cdef RankedCFGEdge ej
	# k1 is the global k
	# first visit of vertex v?
	if label not in cand[start][end]:
		# initialize the heap
		cand[start][end][label] = getcandidatescfg(chart, label, start, end, k1)
	while label not in D[start][end] or len(D[start][end][label]) < k:
		if label in D[start][end]:
			# last derivation
			entry = D[start][end][label][-1]
			ej = entry.key
			# update the heap, adding the successors of last derivation
			lazynextcfg(ej, k1, D, cand, chart, explored)
		# get the next best derivation and delete it from the heap
		if cand[start][end][label]:
			D[start][end].setdefault(label, []).append(
				(<Agenda>cand[start][end][label]).popentry())
		else:
			break
	return D


cdef inline lazynextcfg(RankedCFGEdge ej, int k1, list D, list cand,
		list chart, set explored):
	cdef RankedCFGEdge ej1
	cdef CFGEdge ec = ej.edge
	cdef double prob
	cdef UInt label
	cdef UChar start, end
	# add the |e| neighbors
	# left child
	label = 0 if ec.rule is NULL else ec.rule.rhs1
	start = ej.start
	end = ec.mid
	ej1 = RankedCFGEdge(ej.label, ej.start, ej.end, ej.edge,
			ej.left + 1, ej.right)
	# recursively solve a subproblem
	# NB: increment j1[i] again because j is zero-based and k is not
	lazykthbestcfg(label, start, end, ej1.left + 1, k1,
						D, cand, chart, explored)
	# if it exists and is not in heap yet
	if ((label in D[start][end] and ej1.left < len(D[start][end][label]))
		and ej1 not in explored):  # cand[ej1.head]): <= gives duplicates
		prob = getprobcfg(chart, D, ej1)
		# add it to the heap
		cand[ej1.start][ej1.end][ej1.label][ej1] = prob
		explored.add(ej1)
	# right child?
	if ej.right == -1:
		return
	label = 0 if ec.rule is NULL else ec.rule.rhs2
	start = ec.mid
	end = ej.end
	ej1 = RankedCFGEdge(ej.label, ej.start, ej.end, ej.edge,
			ej.left, ej.right + 1)
	lazykthbestcfg(label, start, end, ej1.right + 1, k1,
						D, cand, chart, explored)
	# if it exists and is not in heap yet
	if ((label in D[start][end] and ej1.right < len(D[start][end][label]))
		and ej1 not in explored):  # cand[ej1.head]): <= gives duplicates
		prob = getprobcfg(chart, D, ej1)
		# add it to the heap
		cand[ej1.start][ej1.end][ej1.label][ej1] = prob
		explored.add(ej1)


cdef inline double getprobcfg(list chart, list D, RankedCFGEdge ej) except -1.:
	""" Get inside probability of 'ej'; try looking in D, or else use chart. """
	cdef CFGEdge ec, edge
	cdef Entry entry
	cdef double result, prob
	ec = ej.edge
	label = 0 if ec.rule is NULL else ec.rule.rhs1
	start = ej.start
	end = ec.mid
	if label in D[start][end]:
		entry = D[start][end][label][ej.left]
		prob = entry.value
	elif ej.left == 0:
		edge = min(chart[start][end][label])
		prob = edge.inside
	else:
		raise ValueError("non-zero rank vector not part of explored derivations")
	# NB: edge.inside if preterminal, 0.0 for terminal
	result = (0.0 if ec.rule is NULL else ec.rule.prob) + prob
	if ej.right >= 0:  # if e.right.label:
		label = 0 if ec.rule is NULL else ec.rule.rhs2
		start = ec.mid
		end = ej.end
		if label in D[start][end]:
			entry = D[start][end][label][ej.right]
			prob = entry.value
		elif ej.right == 0:
			edge = min(chart[start][end][label])
			prob = edge.inside
		else:
			raise ValueError(
					"non-zero rank vector not part of explored derivations")
		result += prob
	return result


cpdef list lazykbestcfg(list chart, CFGChartItem goal, int k):
	""" wrapper function to run lazykthbestcfg.
	does not give actual derivations, but the ranked chart D. """
	cdef Entry entry
	cdef list D = [[{} for _ in x] for x in chart]
	cdef list cand = [[{} for _ in x] for x in chart]
	cdef set explored = set()
	lazykthbestcfg(goal.label, goal.start, goal.end, k, k, D, cand,
			chart, explored)
	return D


cdef inline bint explorederivationcfg(RankedCFGEdge ej,
		list D, list chart, set explored, int n):
	""" Traverse a derivation to ensured RankedEdges are present in D
	for every edge. """
	cdef Entry entry
	cdef RankedCFGEdge rankededge
	cdef UInt label
	if n > 100:
		return False  # hardcoded limit to prevent cycles
	elif ej.edge.rule is NULL:
		return True
	label = ej.edge.rule.rhs1
	if ej.left != -1 and label in chart[ej.start][ej.edge.mid]:
		if label not in D[ej.start][ej.edge.mid]:
			assert ej.left == 0, "non-best edge missing in derivations"
			entry = (<Agenda>getcandidatescfg(chart, label, ej.start,
					ej.edge.mid, 1)).popentry()
			D[ej.start][ej.edge.mid][label] = [entry]
			explored.add(entry.key)
		if not explorederivationcfg(<RankedCFGEdge>
				(<Entry>D[ej.start][ej.edge.mid][label][ej.left]).key,
				D, chart, explored, n + 1):
			return False
	label = ej.edge.rule.rhs2
	if ej.right != -1 and label in chart[ej.edge.mid][ej.end]:
		if label not in D[ej.edge.mid][ej.end]:
			assert ej.right == 0, "non-best edge missing in derivations"
			entry = (<Agenda>getcandidatescfg(chart, label, ej.edge.mid,
					ej.end, 1)).popentry()
			explored.add(entry.key)
			D[ej.edge.mid][ej.end][label] = [entry]
		return explorederivationcfg(<RankedCFGEdge>
				(<Entry>D[ej.edge.mid][ej.end][label][ej.right]).key,
				D, chart, explored, n + 1)
	return True


cdef inline getderivationcfg(result, RankedCFGEdge ej, list  D,
		list chart, list tolabel, bytes debin):
	cdef Entry entry
	cdef RankedCFGEdge rankededge
	cdef UInt label
	if debin is None or debin not in tolabel[ej.label]:
		result += b'('
		result += tolabel[ej.label]
		result += b' '
	if ej.edge.rule is NULL:  # this must be a terminal
		result += str(ej.start).encode('ascii')
	elif ej.left != -1:
		label = ej.edge.rule.rhs1
		rankededge = (<Entry>D[ej.start][ej.edge.mid][label][ej.left]).key
		getderivationcfg(result, rankededge, D, chart, tolabel, debin)
		if ej.right != -1:
			result += b' '
			label = ej.edge.rule.rhs2
			rankededge = (<Entry>D[ej.edge.mid][ej.end][label][ej.right]).key
			getderivationcfg(result, rankededge, D, chart, tolabel, debin)
	if debin is None or debin not in tolabel[ej.label]:
		result += b')'
# --- end CFG specific

cdef bint explorederivation(RankedEdge ej, dict D, dict chart, set explored,
		int n):
	""" Traverse a derivation to ensure RankedEdges are present
	in D for every edge. """
	cdef Entry entry
	cdef RankedEdge rankededge
	if n > 100:
		return False  # hardcoded limit to prevent cycles
	if ej.left != -1 and ej.edge.left in chart:
		if ej.edge.left not in D:
			assert ej.left == 0, "non-best edge missing in derivations"
			entry = (<Agenda>getcandidates(chart, ej.edge.left, 1)).popentry()
			D[ej.edge.left] = [entry]
			explored.add(entry.key)
		if not explorederivation(
				<RankedEdge>(<Entry>D[ej.edge.left][ej.left]).key,
				D, chart, explored, n + 1):
			return False
	if ej.right != -1 and ej.edge.right in chart:
		if ej.edge.right not in D:
			assert ej.right == 0, "non-best edge missing in derivations"
			entry = (<Agenda>getcandidates(chart, ej.edge.right, 1)).popentry()
			D[ej.edge.right] = [entry]
			explored.add(entry.key)
		return explorederivation(
				<RankedEdge>(<Entry>D[ej.edge.right][ej.right]).key,
				D, chart, explored, n + 1)
	return True


cdef inline getderivationlcfrs(result, RankedEdge ej, dict D,
		dict chart, list tolabel, bytes debin):
	cdef Entry entry
	cdef RankedEdge rankededge
	cdef ChartItem item
	if debin is None or debin not in tolabel[ej.head.label]:
		result += b'('
		result += tolabel[ej.head.label]
		result += b' '
	item = ej.edge.left
	if item not in chart:
		# this must be a terminal
		result += str(item.lexidx()).encode('ascii')
	else:
		rankededge = (<Entry>D[item][ej.left]).key
		getderivationlcfrs(result, rankededge, D, chart, tolabel, debin)
	if ej.right != -1:
		item = ej.edge.right
		result += b' '
		if item in chart:
			rankededge = (<Entry>D[item][ej.right]).key
			getderivationlcfrs(result, rankededge, D, chart, tolabel, debin)
		else:
			result += str(item.lexidx()).encode('ascii')
	if debin is None or debin not in tolabel[ej.head.label]:
		result += b')'


def getderiv(ej, D, chart, list tolabel, bytes debin):
	""" Translate the (e, j) notation to an actual tree string in
	bracket notation.  e is an edge, j is a vector prescribing the rank of the
	corresponding tail node. For example, given the edge <S, [NP, VP], 1.0> and
	vector [2, 1], this points to the derivation headed by S and having the 2nd
	best NP and the 1st best VP as children.
	If `debin` is specified, will perform on-the-fly debinarization of nodes
	with labels containing `debin` an a substring. """
	result = bytearray()
	if isinstance(ej, RankedEdge):
		getderivationlcfrs(result, ej, D, chart, tolabel, debin)
	elif isinstance(ej, RankedCFGEdge):
		getderivationcfg(result, ej, D, chart, tolabel, debin)
	return str(result.decode('ascii'))


cpdef tuple lazykbest(chart, ChartItem goal, int k, list tolabel=None,
		bytes debin=None, bint derivs=True):
	""" wrapper function to run lazykthbest and get the actual derivations,
	(except when derivs is False) as well as the ranked chart.
	chart is a monotone hypergraph; should be acyclic unless probabilities
	resolve the cycles (maybe nonzero weights for unary productions are
	sufficient?).
	maps ChartItems to lists of tuples with ChartItems and a weight. The
	items in each list are to be ordered as they were added by the viterbi
	parse, with the best item first.
	goal is a ChartItem that is to be the root node of the derivations.
	k is the number of derivations desired.
	tolabel is a dictionary mapping numeric IDs to the original nonterminal
	labels.  """
	cdef Entry entry
	cdef set explored = set()
	derivations = []
	if isinstance(goal, CFGChartItem):
		D = [[{} for _ in x] for x in chart]
		cand = [[{} for _ in x] for x in chart]
		start = (<CFGChartItem>goal).start
		end = (<CFGChartItem>goal).end
		lazykthbestcfg(goal.label, start, end, k, k, D, cand, chart, explored)
		D[start][end][goal.label] = [entry
				for entry in D[start][end][goal.label][:k]
				if explorederivationcfg(entry.key, D, chart, explored, 0)]
		if derivs:
			derivations = [(getderiv(
					entry.key, D, chart, tolabel, debin), entry.value)
					for entry in D[start][end][goal.label]]
	else:
		D = {}
		cand = {}
		lazykthbest(goal, k, k, D, cand, chart, explored)
		D[goal] = [entry for entry in D[goal][:k]
				if explorederivation(entry.key, D, chart, explored, 0)]
		if derivs:
			derivations = [(getderiv(
					entry.key, D, chart, tolabel, debin), entry.value)
					for entry in D[goal]]
	return derivations, D, explored


cpdef main():
	""" Simple demonstration. """
	from math import log, exp
	from pprint import pprint
	cdef SmallChartItem v, ci
	cdef LCFRSEdge ed
	cdef RankedEdge re
	cdef Entry entry
	cdef Rule rules[11]
	tolabel = b"Epsilon S NP V ADV VP VP2 PN".split()
	toid = {a: n for n, a in enumerate(tolabel)}
	NONE = (b"Epsilon", 0)			# sentinel node
	chart = {
			(b"S", 0b111): [
				((0.7 * 0.9 * 0.5), 0.7,
						(b"NP", 0b100), (b"VP2", 0b011)),
				((0.4 * 0.9 * 0.5), 0.4,
						(b"NP", 0b100), (b"VP", 0b011))],
			(b"VP", 0b011): [
				(0.5, 0.5, (b"V", 0b010), (b"ADV", 0b001)),
				(0.4, 0.4, ("walks", 1), (b"ADV", 0b001))],
			(b"VP2", 0b011): [
				(0.5, 0.5, (b"V", 0b010), (b"ADV", 0b001)),
				(0.4, 0.4, ("walks", 1), (b"ADV", 0b001))],
			(b"NP", 0b100): [(0.5, 0.5, ("Mary", 0), NONE),
							(0.9, 0.9, (b"PN", 0b100), NONE)],
			(b"PN", 0b100): [(1.0, 1.0, ("Mary", 0), NONE),
							(0.9, 0.9, (b"NP", 0b100), NONE)],
			(b"V", 0b010): [(1.0, 1.0, ("walks", 1), NONE)],
			(b"ADV", 0b001): [(1.0, 1.0, ("quickly", 2), NONE)]
		}
	# a hack to make Rule structs with the right probabilities.
	# rules[7] will be a Rule with probability 0.7
	for a in range(1, 11):
		rules[a].prob = -log(a / 10.0)
	for a in list(chart):
		chart[SmallChartItem(toid[a[0]], a[1])] = {x: x
			for x in [new_LCFRSEdge(-log(c), -log(c), &(rules[int(d * 10)]),
			SmallChartItem(toid.get(e, 0), f),
			SmallChartItem(toid.get(g, 0), h))
			for c, d, (e, f), (g, h) in chart.pop(a)]}
	assert SmallChartItem(toid[b"NP"], 0b100) == SmallChartItem(
			toid[b"NP"], 0b100)
	cand = {}
	D = {}
	k = 10
	goal = SmallChartItem(toid[b"S"], 0b111)
	for v, b in lazykthbest(goal, k, k, D, cand, chart, set()).items():
		print(tolabel[v.label].decode('ascii'), bin(v.vec)[2:])
		for entry in b:
			re = entry.key
			ed = re.edge
			j = (re.left, )
			if re.right != -1:
				j += (re.right, )
			ip = entry.value
			print(tolabel[v.label].decode('ascii'), ":",
				' '.join([tolabel[ci.label].decode('ascii') for ci, _
				in zip((ed.left, ed.right), j)]),
				exp(-ed.rule.prob), j, exp(-ip))
		print()
	print("tolabel", end='')
	pprint(tolabel)
	print("candidates", end='')
	for a in cand:
		print(a, len(cand[a]), end='')
		pprint(list(cand[a].items()))

	print("\n%d derivations" % (len(D[goal])))
	derivations = lazykbest(chart, goal, k, tolabel)[0]
	for a, p in derivations:
		print(exp(-p), a)
	assert len(D[goal]) == len(set(D[goal]))
	assert len(derivations) == len(set(derivations))
	assert len(set(derivations)) == len(dict(derivations))

if __name__ == '__main__':
	main()



""" Read and write treebanks. """
from __future__ import division, print_function, unicode_literals
import io
import os
import re
import xml.etree.cElementTree as ElementTree
from glob import glob
from itertools import count, chain
from collections import defaultdict, OrderedDict, Counter as multiset
from operator import itemgetter
from discodop.tree import Tree, ParentedTree
from discodop.treetransforms import addbitsets, fanout
from discodop.treebanktransforms import punctremove, punctraise, \
		balancedpunctraise, punctroot, ispunct

FIELDS = tuple(range(6))
WORD, LEMMA, TAG, MORPH, FUNC, PARENT = FIELDS
POSRE = re.compile(r"\(([^() ]+) [^ ()]+\)")
TERMINALSRE = re.compile(r" ([^ ()]+)\)")
EXPORTNONTERMINAL = re.compile(r"^#([0-9]+)$")
LEAVESRE = re.compile(r" ([^ ()]*)\)")
FRONTIERNTRE = re.compile(r" \)")


class CorpusReader(object):
	""" Abstract corpus reader. """
	def __init__(self, root, fileids, encoding='utf-8', headrules=None,
				headfinal=True, headreverse=False, markheads=False, punct=None,
				functions=None, morphology=None, lemmas=None):
		"""
		:param headrules: if given, read rules for assigning heads and apply
			them by ordering constituents according to their heads
		:param headfinal: whether to put the head in final or in frontal
			position
		:param headreverse: the head is made final/frontal by reversing
			everything before or after the head. When true, the side on which
			the head is will be the reversed side.
		:param markheads: add '^' to phrasal label of heads.
		:param punct: one of ...
			:None: leave punctuation as is.
			:'move': move punctuation to appropriate constituents
					using heuristics.
			:'remove': eliminate punctuation.
			:'root': attach punctuation directly to root
					(as in original Negra/Tiger treebanks).
		:param functions: one of ...
			:None: leave syntactic labels as is.
			:'add': concatenate grammatical function to syntactic label,
				separated by a hypen: e.g., NP => NP-SBJ
			:'remove': strip away hyphen-separated grammatical function,
				e.g., NP-SBJ => NP
			:'replace': replace syntactic label with grammatical function,
				e.g., NP => SBJ
		:param morphology: one of ...
			:None: use POS tags as preterminals
			:'add': concatenate morphological information to POS tags,
				e.g., DET/sg.def
			:'replace': use morphological information as preterminal label
			:'between': add node with morphological information between
				POS tag and word, e.g., (DET (sg.def the))
		:param lemmas: one of ...
			:None: ignore lemmas
			:'between': insert lemma as node between POS tag and word. """
		self.reverse = headreverse
		self.headfinal = headfinal
		self.markheads = markheads
		self.functions = functions
		self.punct = punct
		self.morphology = morphology
		self.lemmas = lemmas
		self.headrules = readheadrules(headrules) if headrules else {}
		self._encoding = encoding
		if fileids == '':
			fileids = '*'
		self._filenames = sorted(glob(os.path.join(root, fileids)), key=numbase)
		assert functions in (None, 'leave', 'add', 'remove', 'replace'), (
				functions)
		assert punct in (None, 'move', 'remove', 'root')
		assert self._filenames, (
				"no files matched pattern %s" % os.path.join(root, fileids))
		self._sents_cache = None
		self._tagged_sents_cache = None
		self._parsed_sents_cache = None
		self._block_cache = self._read_blocks()

	def parsed_sents(self):
		""" :returns: an ordered dictionary of parse trees (``Tree`` objects \
		with integer indices as leaves). """
		if not self._parsed_sents_cache:
			self._parsed_sents_cache = OrderedDict((a, self._parsetree(b))
					for a, b in self._block_cache.items())
		return self._parsed_sents_cache

	def sents(self):
		""" :returns: an ordered dictionary of sentences, \
		each sentence being a list of words. """
		if not self._sents_cache:
			self._sents_cache = OrderedDict((a, self._word(b))
					for a, b in self._block_cache.items())
		return self._sents_cache

	def tagged_sents(self):
		""" :returns: an ordered dictionary of tagged sentences, \
		each tagged sentence being a list of (word, tag) pairs. """
		if not self._tagged_sents_cache:
			# for each sentence, zip its words & tags together in a list.
			# this assumes that .sents() and .parsed_sents() correctly remove
			# punctuation if requested.
			self._tagged_sents_cache = OrderedDict((n,
				list(zip(sent, map(itemgetter(1), sorted(tree.pos())))))
				for (n, sent), tree in zip(self.sents().items(),
						self.parsed_sents().values()))
		return self._tagged_sents_cache

	def blocks(self):
		""" :returns: a list of strings containing the raw representation of \
		trees in the original treebank. """

	def _read_blocks(self):
		""" No-op. For line-oriented formats re-reading is cheaper than
		caching. """

	def _parse(self, block):
		""" :returns: a parse tree given a string from the treebank file. """

	def _parsetree(self, block):
		""" :returns: a transformed parse tree. """
		tree, sent = self._parse(block)
		if not sent:
			return tree
		# roughly order constituents by order in sentence
		for a in reversed(list(tree.subtrees(lambda x: len(x) > 1))):
			a.sort(key=Tree.leaves)
		if self.punct == "remove":
			punctremove(tree, sent)
		elif self.punct == "move":
			punctraise(tree, sent)
			balancedpunctraise(tree, sent)
			# restore order
			for a in reversed(list(tree.subtrees(lambda x: len(x) > 1))):
				a.sort(key=Tree.leaves)
		elif self.punct == "root":
			punctroot(tree, sent)
		if self.headrules:
			for node in tree.subtrees(lambda n: n and isinstance(n[0], Tree)):
				sethead(headfinder(node, self.headrules))
				headorder(node, self.headfinal, self.reverse)
				if self.markheads:
					headmark(node)
		handlefunctions(self.functions, tree)
		return tree

	def _word(self, block, orig=False):
		""" :returns: a list of words given a string.
		When orig is True, return original sentence verbatim;
		otherwise it will follow parameters for punctuation. """


class NegraCorpusReader(CorpusReader):
	""" Read a corpus in the Negra export format. """
	def blocks(self):
		return OrderedDict((a, "#BOS %s\n%s\n#EOS %s\n" % (a,
				"\n".join("\t".join(c) for c in b), a))
				for a, b in self._block_cache.items())

	def _read_blocks(self):
		""" Read corpus and return list of blocks corresponding to each
		sentence. """
		result = OrderedDict()
		started = False
		for filename in self._filenames:
			for line in io.open(filename, encoding=self._encoding):
				if line.startswith('#BOS '):
					assert not started, ("beginning of sentence marker while "
							"previous one still open: %s" % line)
					started = True
					sentid = line.strip().split()[1]
					lines = []
				elif line.startswith('#EOS '):
					assert started, "end of sentence marker while none started"
					thissentid = line.strip().split()[1]
					assert sentid == thissentid, ("unexpected sentence id: "
							"start=%s, end=%s" % (sentid, thissentid))
					started = False
					assert sentid not in result, (
							"duplicate sentence ID: %s" % sentid)
					result[sentid] = lines
				elif started:
					lines.append(exportsplit(line))
		return result

	def _parse(self, block):
		tree = exportparse(block, self.morphology, self.lemmas)
		sent = self._word(block, orig=True)
		return tree, sent

	def _word(self, block, orig=False):
		if orig or self.punct != "remove":
			return [a[WORD] for a in block
					if not EXPORTNONTERMINAL.match(a[WORD])]
		return [a[WORD] for a in block if not EXPORTNONTERMINAL.match(a[WORD])
			and not ispunct(a[WORD], a[TAG])]


class DiscBracketCorpusReader(CorpusReader):
	""" A corpus reader where the phrase-structure is represented by a tree in
	bracket notation, where the leaves are indices pointing to words in a
	separately represented sentence; e.g.:
	(S (NP 1) (VP (VB 0) (JJ 2)) (? 3))	is John rich ?
	Note that the tree and the sentence are separated by a tab, while the words
	in the sentence are separated by spaces. There is one tree and sentence
	per line.
	Compared to Negra's export format, this format lacks morphology, lemmas and
	functional edges. On the other hand, it is very close to the internal
	representation employed here, so it can be read efficiently. """
	def sents(self):
		return OrderedDict((n, self._word(line))
				for n, line in self.blocks().items())

	def parsed_sents(self):
		if not self._parsed_sents_cache:
			self._parsed_sents_cache = OrderedDict(enumerate(
					map(self._parsetree, self.blocks().values()), 1))
		return self._parsed_sents_cache

	def blocks(self):
		return OrderedDict(enumerate(filter(None,
			(line for filename in self._filenames
			for line in io.open(filename, encoding=self._encoding))), 1))

	def _parse(self, block):
		result = ParentedTree.parse(block.split("\t", 1)[0], parse_leaf=int)
		sent = self._word(block, orig=True)
		return result, sent

	def _word(self, block, orig=False):
		sent = block.split("\t", 1)[1].rstrip("\n\r").split(' ')
		if orig or self.punct != "remove":
			return sent
		return [a for a in sent if not ispunct(a, None)]


class BracketCorpusReader(CorpusReader):
	""" A standard corpus reader where the phrase-structure is represented by a
	tree in bracket notation; e.g.:
	(S (NP John) (VP (VB is) (JJ rich)) (. .))
	TODO: support traces & empty nodes. """
	def sents(self):
		if not self._sents_cache:
			self._sents_cache = OrderedDict((n, self._word(tree))
				for n, tree in self.blocks().items())
		return self._sents_cache

	def tagged_sents(self):
		if not self._tagged_sents_cache:
			self._tagged_sents_cache = OrderedDict(
				(n, list(zip(sent, POSRE.findall(block)))) for (n, sent), block
				in zip(self.sents().items(), self.blocks().values()))
		return self._tagged_sents_cache

	def parsed_sents(self):
		if not self._parsed_sents_cache:
			self._parsed_sents_cache = OrderedDict(enumerate(
					map(self._parsetree, self.blocks().values()), 1))
		return self._parsed_sents_cache

	def blocks(self):
		return OrderedDict(enumerate((line for filename in self._filenames
			for line in io.open(filename, encoding=self._encoding) if line), 1))

	def _parse(self, block):
		c = count()
		result = ParentedTree.parse(block, parse_leaf=lambda _: next(c))
		if result.label not in ('TOP', 'ROOT'):
			result = ParentedTree('TOP', [result])
		sent = self._word(block, orig=True)
		return result, sent

	def _word(self, block, orig=False):
		sent = TERMINALSRE.findall(block)
		if orig or self.punct != "remove":
			return sent
		return [a for a in sent if not ispunct(a, None)]


class AlpinoCorpusReader(CorpusReader):
	""" Corpus reader for the Dutch Alpino treebank in XML format. """

	def blocks(self):
		""" :returns: a list of strings containing the raw representation of \
		trees in the treebank. """
		if self._block_cache is None:
			self._block_cache = self._read_blocks()
		return OrderedDict((n, ElementTree.tostring(a, encoding='UTF-8'))
				for n, a in self._block_cache.items())

	def _read_blocks(self):
		""" Read corpus and return list of blocks corresponding to each
		sentence. """
		results = OrderedDict()
		assert self._encoding in (None, 'utf8', 'utf-8'), (
				"Encoding specified in XML files.")
		for filename in self._filenames:
			block = ElementTree.parse(filename).getroot()
			#n = s.find('comments')[0].text.split('|', 1)[0], s
			# ../path/dir/file.xml => dir/file
			path, filename = os.path.split(filename)
			_, lastdir = os.path.split(path)
			n = os.path.join(lastdir, filename)[:-len('.xml')]
			results[n] = block
		return results

	def _parse(self, block):
		""" :returns: a parse tree given a string. """
		def getsubtree(node, parent):
			""" Traverse Alpino XML tree and create Tree object. """
			# FIXME: proper representation for arbitrary features
			source = [''] * len(FIELDS)
			source[WORD] = node.get('word') or ("#%s" % node.get('id'))
			source[LEMMA] = node.get('lemma') or node.get('root')
			source[MORPH] = node.get('postag') or node.get('frame')
			source[FUNC] = node.get('rel')
			if 'cat' in node.keys():
				source[TAG] = node.get('cat')
				if node.get('index'):
					coindexed[node.get('index')] = source
				label = node.get('cat')
				result = ParentedTree(label.upper(), [])
				for child in node:
					subtree = getsubtree(child, result)
					if subtree and (
							'word' in child.keys() or 'cat' in child.keys()):
						subtree.source[PARENT] = node.get('id')
						result.append(subtree)
				if not len(result):
					return None
			elif 'word' in node.keys():
				source[TAG] = node.get('pt') or node.get('pos')
				if node.get('index'):
					coindexed[node.get('index')] = source
				result = ParentedTree('', list(
						range(int(node.get('begin')), int(node.get('end')))))
				handlemorphology(self.morphology, self.lemmas, result, source)
			elif 'index' in node.keys():
				coindexation[node.get('index')].extend(
						(node.get('rel'), parent))
				return None
			result.source = source
			return result
		coindexed = {}
		coindexation = defaultdict(list)
		# NB: in contrast to Negra export format, don't need to add
		# root/top node
		result = getsubtree(block.find('node'), None)
		# FIXME: need MultipleParentedTree for secedges
		#for index, secedges in coindexation.items():
		#	coindexed[index].extend(secedges)
		sent = self._word(block)
		return result, sent

	def _word(self, block, orig=False):
		if orig or self.punct != "remove":
			return block.find('sentence').text.split()
		return [word for word in block.find('sentence').text.split()
				if not ispunct(word, None)]  # fixme: don't have tag


class TigerXMLCorpusReader(CorpusReader):
	""" Corpus reader for the Tiger XML format. """
	def blocks(self):
		""" :returns: a list of strings containing the raw representation of \
		trees in the treebank. """
		if self._block_cache is None:
			self._block_cache = self._read_blocks()
		return OrderedDict((n, ElementTree.tostring(a, encoding='UTF-8'))
				for n, a in self._block_cache.items())

	def _read_blocks(self):
		results = OrderedDict()
		for filename in self._filenames:
			# todo: use iterparse()
			block = ElementTree.parse(filename).getroot()
			for sent in block.find('body').findall('s'):
				results[sent.get('id')] = sent
		return results

	def _parse(self, block):
		""" Translate Tiger XML structure to the fields of export format,
		so that tree writing code for export format can be used. """
		nodes = OrderedDict()
		root = block.find('graph').get('root')
		for term in block.find('graph').find('terminals'):
			fields = nodes.setdefault(term.get('id'), 6 * [None])
			fields[WORD] = term.get('word')
			fields[LEMMA] = term.get('lemma')
			fields[TAG] = term.get('pos')
			fields[MORPH] = term.get('morph')
			nodes[term.get('id')] = fields
		for nt in block.find('graph').find('nonterminals'):
			if nt.get('id') == root:
				ntid = '0'
			else:
				fields = nodes.setdefault(nt.get('id'), 6 * [None])
				ntid = nt.get('id').split('_')[-1]
				fields[WORD] = '#' + ntid
				fields[TAG] = nt.get('cat')
				fields[LEMMA] = fields[MORPH] = '--'
			for edge in nt:
				idref = edge.get('idref')
				nodes.setdefault(idref, 6 * [None])
				if nodes[idref][FUNC] is None:
					nodes[idref][FUNC] = edge.get('label')
					nodes[idref][PARENT] = ntid
				else:  # secondary edge
					nodes[idref].extend((edge.get('label'), ntid))
		tree = exportparse(list(nodes.values()), self.morphology, self.lemmas)
		sent = self._word(block, orig=True)
		return tree, sent

	def _word(self, block, orig=False):
		if orig or self.punct != "remove":
			return [term.get('word')
				for term in block.find('graph').find('terminals')]
		return [term.get('word')
			for term in block.find('graph').find('terminals')
			if not ispunct(term.get('word'), term.get('pos'))]


def numbase(key):
	""" Turn a file name into a numeric sorting key if possible. """
	path, base = os.path.split(key)
	base = base.split(".", 1)
	try:
		base[0] = int(base[0])
	except ValueError:
		pass
	return [path] + base


def exportsplit(line):
	""" take a line in export format and split into fields,
	add dummy fields lemma, sec. edge if those fields are absent. """
	if "%%" in line:  # we don't want comments.
		line = line[:line.index("%%")]
	fields = line.split()
	fieldlen = len(fields)
	if fieldlen == 5:
		fields[1:1] = ['']
		fields.extend(['', ''])
	elif fieldlen == 6:
		fields.extend(['', ''])
	elif fieldlen < 8 or fieldlen & 1:
		# NB: zero or more sec. edges come in pairs of parent id and label
		raise ValueError(
				'expected 5 or 6+ even number of columns: %r' % fields)
	return fields


def exportparse(block, morphology=None, lemmas=None):
	""" Given a tree in export format as a list of lists,
	construct a Tree object for it. """
	def getchildren(parent):
		""" Traverse tree in export format and create Tree object. """
		results = []
		for n, source in children[parent]:
			# n is the index in the block to record word indices
			m = EXPORTNONTERMINAL.match(source[WORD])
			if m:
				child = ParentedTree(source[TAG], getchildren(m.group(1)))
			else:  # POS + terminal
				child = ParentedTree('', [n])
				handlemorphology(morphology, lemmas, child, source)
			child.source = tuple(source)
			results.append(child)
		return results

	children = {}
	for n, source in enumerate(block):
		children.setdefault(source[PARENT], []).append((n, source))
	result = ParentedTree('ROOT', getchildren('0'))
	return result


def getreader(fmt):
	""" :returns: the corpus reader class with name ``fmt``. """
	return {'export': NegraCorpusReader,
		'discbracket': DiscBracketCorpusReader,
		'bracket': BracketCorpusReader,
		'alpino': AlpinoCorpusReader}[fmt]


def splitpath(path):
	""" Split path into a pair of (directory, filename). """
	return path.rsplit('/', 1) if '/' in path else ('.', path)


indexre = re.compile(r" [0-9]+\)")


def writetree(tree, sent, n, fmt, headrules=None, morph=None):
	""" Convert a tree with indices as leafs and a sentence with the
	corresponding non-terminals to a single string in the given format.
	Formats are bracket, discbracket, and Negra's export format,
	as well unlabelled dependency conversion into mst or conll format
	(requires head rules). Lemmas, functions, and morphology information will
	be empty unless nodes contain a 'source' attribute with such information.
	"""
	def getword(idx):
		""" Get word given an index and restore parentheses. """
		return sent[int(idx[:-1])].replace('(', '-LRB-').replace(')', '-RRB-')
	if fmt == "alpino":
		fmt = "export"  # FIXME implement Alpino XML output?

	if fmt == "bracket":
		return indexre.sub(lambda x: ' %s)' % getword(x.group()),
				"%s\n" % tree)
	elif fmt == "discbracket":
		return "%s\t%s\n" % (tree, ' '.join(sent))
	elif fmt == "export":
		result = []
		if n is not None:
			result.append("#BOS %s" % n)
		indices = tree.treepositions('leaves')
		wordsandpreterminals = indices + [a[:-1] for a in indices]
		phrasalnodes = [a for a in tree.treepositions('postorder')
				if a not in wordsandpreterminals and a != ()]
		wordids = {tree[a]: a for a in indices}
		assert len(sent) == len(indices) == len(wordids), (sent, wordids.keys())
		for i, word in enumerate(sent):
			assert word, 'empty word in sentence: %r' % sent
			idx = wordids[i]
			node = tree[idx[:-1]]
			lemma = '--'
			postag = node.label.replace('$[', '$(') or '--'
			func = morph = '--'
			secedges = []
			if getattr(node, 'source', None):
				lemma = node.source[LEMMA] or '--'
				morph = node.source[MORPH] or '--'
				func = node.source[FUNC] or '--'
				secedges = node.source[6:]
			if morph == '--':
				morph = node.label if morph == 'replace' else '--'
			nodeid = str(500 + phrasalnodes.index(idx[:-2])
					if len(idx) > 2 else 0)
			result.append("\t".join((word, lemma, postag, morph, func, nodeid)
					+ tuple(secedges)))
		for idx in phrasalnodes:
			node = tree[idx]
			parent = '#%d' % (500 + phrasalnodes.index(idx))
			lemma = '--'
			label = node.label or '--'
			func = morph = '--'
			secedges = []
			if getattr(node, 'source', None):
				morph = node.source[MORPH] or '--'
				func = node.source[FUNC] or '--'
				secedges = node.source[6:]
			nodeid = str(500 + phrasalnodes.index(idx[:-1])
					if len(idx) > 1 else 0)
			result.append('\t'.join((parent, lemma, label, morph, func, nodeid)
					+ tuple(secedges)))
		if n is not None:
			result.append("#EOS %s" % n)
		return "%s\n" % "\n".join(result)
	elif fmt in ("conll", "mst"):
		assert headrules, "dependency conversion requires head rules."
		deps = dependencies(tree, headrules)
		if fmt == "mst":  # MST parser can read this format
			# fourth line with function tags is left empty.
			return "\n".join((
				"\t".join(word for word in sent),
				"\t".join(tag for _, tag in sorted(tree.pos())),
				"\t".join(str(head) for _, _, head in deps))) + "\n\n"
		elif fmt == "conll":
			return "\n".join("%d\t%s\t_\t%s\t%s\t_\t%d\t%s\t_\t_" % (
				n, word, tag, tag, head, rel)
				for word, (_, tag), (rel, n, head)
				in zip(sent, sorted(tree.pos()), deps)) + "\n\n"
	else:
		raise ValueError("unrecognized format: %r" % fmt)


def handlefunctions(action, tree, pos=True, top=False):
	""" Add function tags to phrasal labels e.g., 'VP' => 'VP-HD'.

	:param action: one of {None, 'add', 'replace', 'remove'}
	:param pos: whether to add function tags to POS tags.
	:param top: whether to add function tags to the top node. """
	if action in (None, 'leave'):
		return
	for a in tree.subtrees():
		if action == 'remove':
			# e.g., NP-SUBJ or NP=2 => NP, but don't touch -NONE-
			x = a.label.find('-')
			if x > 0:
				a.label = a.label[:x]
			x = a.label.find('=')
			if x > 0:
				a.label = a.label[:x]
		else:
			if not top and a is tree:  # skip TOP label
				continue
			if pos or isinstance(a[0], Tree):
				# test for non-empty function tag ("---" is considered empty)
				if (getattr(a, 'source', None)
						and any(a.source[FUNC].split("-"))):
					func = a.source[FUNC].split("-")[0].upper()
					if action == 'add':
						a.label += "-%s" % func
					elif action == 'replace':
						a.label = func


def handlemorphology(action, lemmaaction, preterminal, source):
	""" Given a preterminal, augment or replace its label with morphological
	information. """
	# escape any parentheses to avoid hassles w/bracket notation of trees
	tag = source[TAG].replace('(', '[').replace(')', ']')
	morph = source[MORPH].replace('(', '[').replace(')', ']')
	lemma = source[LEMMA].replace('(', '[').replace(')', ']') or '--'
	if lemmaaction == 'add':
		raise NotImplementedError
		#sent[preterminal[0]] = '%s|%s' % (word, lemma)
	elif lemmaaction == 'replace':
		raise NotImplementedError
		#sent[preterminal[0]] = lemma
	elif lemmaaction == 'between':
		preterminal[:] = [preterminal.__class__(lemma, preterminal)]
	elif lemmaaction not in (None, 'no'):
		raise ValueError

	if action in (None, 'no'):
		preterminal.label = tag
	elif action == 'add':
		preterminal.label = '%s/%s' % (tag, morph)
	elif action == 'replace':
		preterminal.label = morph
	elif action == 'between':
		preterminal[:] = [preterminal.__class__(morph, [preterminal.pop()])]
		preterminal.label = tag
	else:
		raise ValueError
	return preterminal


def readheadrules(filename):
	""" Read a file containing heuristic rules for head assigment.
	The file containing head assignment rules for negra is part of rparse,
	under src/de/tuebingen/rparse/treebank/constituent/negra/
	Example line: "s right-to-left vmfin vafin vaimp", which means
	traverse siblings of an S constituent from right to left, the first child
	with a label of vmfin, vafin, or vaimp will be marked as head. """
	headrules = {}
	for line in open(filename):
		line = line.strip().upper()
		if line and not line.startswith("%") and len(line.split()) > 2:
			label, lr, heads = line.split(None, 2)
			headrules.setdefault(label, []).append((lr, heads.split()))
	return headrules


def headfinder(tree, headrules, headlabels=frozenset({'HD'})):
	""" use head finding rules to select one child of tree as head. """
	candidates = [a for a in tree if getattr(a, 'source', None)
			and headlabels.intersection(a.source[FUNC].upper().split('-'))]
	if candidates:
		return candidates[0]
	for lr, heads in headrules.get(tree.label, []):
		if lr == 'LEFT-TO-RIGHT':
			children = tree
		elif lr == 'RIGHT-TO-LEFT':
			children = tree[::-1]
		else:
			raise ValueError
		for head in heads:
			for child in children:
				if (isinstance(child, Tree)
						and child.label.split('[')[0] == head):
					return child
	# default head is initial nonterminal
	for child in tree:
		if isinstance(child, Tree):
			return child


def sethead(child):
	""" mark node as head in an auxiliary field. """
	child.source = getattr(child, "source", 6 * [''])
	if 'HD' not in child.source[FUNC].upper().split("-"):
		x = list(child.source)
		if child.source[FUNC] in (None, '', '--'):
			x[FUNC] = '-HD'
		else:
			x[FUNC] = x[FUNC] + '-HD'
		child.source = tuple(x)


def headmark(tree):
	""" add marker to label of head node. """
	head = [a for a in tree if getattr(a, 'source', None)
			and 'HD' in a.source[FUNC].upper().split('-')]
	if not head:
		return
	head[-1].label += '-HD'


def headorder(tree, headfinal, reverse):
	""" change order of constituents based on head (identified with
	function tag). """
	head = [n for n, a in enumerate(tree)
		if getattr(a, 'source', None)
		and 'HD' in a.source[FUNC].upper().split("-")]
	if not head:
		return
	headidx = head.pop()
	# everything until the head is reversed and prepended to the rest,
	# leaving the head as the first element
	nodes = tree[:]
	tree[:] = []
	if headfinal:
		if reverse:
			# head final, reverse rhs: A B C^ D E => A B E D C^
			tree[:] = nodes[:headidx] + nodes[headidx:][::-1]
		else:
			# head final, reverse lhs:  A B C^ D E => E D A B C^
			tree[:] = nodes[headidx + 1:][::-1] + nodes[:headidx + 1]
	else:
		if reverse:
			# head first, reverse lhs: A B C^ D E => C^ B A D E
			tree[:] = nodes[:headidx + 1][::-1] + nodes[headidx + 1:]
		else:
			# head first, reverse rhs: A B C^ D E => C^ D E B A
			tree[:] = nodes[headidx:] + nodes[:headidx][::-1]


def saveheads(tree, tailmarker):
	""" When a head-outward binarization is used, this function ensures the
	head is known when the tree is converted to export format. """
	if not tailmarker:
		return
	for node in tree.subtrees(lambda n: tailmarker in n.label):
		node.source = ['--'] * 6
		node.source[FUNC] = 'HD'


def headstats(trees):
	""" collects some information useful for writing headrules. """
	heads = defaultdict(multiset)
	pos1 = defaultdict(multiset)
	pos2 = defaultdict(multiset)
	pos3 = defaultdict(multiset)
	unknown = defaultdict(multiset)
	for tree in trees:
		for a in tree.subtrees(lambda x: len(x) > 1):
			for n, b in enumerate(a):
				if 'hd' in b.source[FUNC]:
					heads[a.label][b.label] += 1
					pos1[a.label][n] += 1
					pos2[a.label][len(a) - (n + 2)] += 1
					pos3[a.label][len(a)] += 1
					break
			else:
				unknown[a.label].update(b.label for b in a)
	return heads, unknown, pos1, pos2, pos3


def dependencies(root, headrules):
	""" Lin (1995): A Dependency-based Method
	for Evaluating Broad-Coverage Parser """
	deps = []
	deps.append(("ROOT", makedep(root, deps, headrules), 0))
	return sorted(deps)


def makedep(root, deps, headrules):
	""" Traverse a tree marking heads and extracting dependencies. """
	if not isinstance(root[0], Tree):
		return root[0] + 1
	headchild = headfinder(root, headrules)
	lexhead = makedep(headchild, deps, headrules)
	for child in root:
		if child is headchild:
			continue
		lexheadofchild = makedep(child, deps, headrules)
		deps.append(("NONE", lexheadofchild, lexhead))
	return lexhead


def incrementaltreereader(treeinput, morphology=None, functions=None):
	""" Incremental corpus reader support brackets, discbrackets, and export
	format. The format is autodetected. Expects an iterator giving one line at
	a time. Yields tuples with a Tree object and a separate lists of terminals.
	"""
	treeinput = chain(iter(treeinput), ('(', None))  # hack
	line = next(treeinput)
	# try the following readers; on the first match the others are dropped.
	readers = [segmentbrackets('()'), segmentbrackets('[]'),
			segmentexport(morphology, functions)]
	for reader in readers:
		reader.send(None)
	x = -1
	while True:
		# status 0: line not part of tree; status 1: waiting for end of tree.
		res, status = None, 1
		for n, reader in enumerate(readers if x == -1 else readers[x:x + 1]):
			while res is None:
				res, status = reader.send(line)
				if status == 0:
					# do not give this line to this reader anymore.
					break
				line = next(treeinput)
			if res is not None:
				if x == -1:
					x = n
				for tree, sent in res:
					yield tree, sent
				break
		if line is None:  # this was the last line
			return
		if res is None:  # none of the readers accepted this line
			line = next(treeinput)


def segmentbrackets(brackets):
	""" Co-routine that accepts one line at a time;
	yields tuples (result, status) where ...

	- result is None or one or more S-expressions as a list of
		tuples (tree, rest), where rest is the string outside of brackets
		between this S-expression and the next.
	- status is 1 if the line was consumed, else 0. """
	lb, rb = brackets
	strtermre = re.compile('[^0-9\\%s]\\%s' % (rb, rb))
	parens = 0  # number of open parens
	prev = ''  # pass on if tree is not yet complete in current line
	result = ''  # tree as string
	results = []  # trees found in current line
	line = (yield None, 1)
	while True:
		start = 0  # index where current tree starts
		a, b = line.find(lb, len(prev)), line.find(rb, len(prev))
		prev = line
		while a != -1 or b != -1:
			if a != -1 and (a < b or b == -1):  # left bracket
				if parens == 0:
					rest, prev = line[start:a], line[a:]
					if result:
						results.append(
								brackettree(result, rest, brackets, strtermre))
						result = ''
						start = a
				parens += 1
				a = line.find(lb, a + 1)
			elif b != -1 and (b < a or a == -1):  # right bracket
				parens -= 1
				if parens == 0:
					result, prev = line[start:b + 1], line[b + 1:]
					start = b + 1
				elif parens < 0:
					#raise ValueError('unbalanced parentheses')
					parens = 0
				b = line.find(rb, b + 1)
		status = 1 if results or result or parens else 0
		line = (yield results or None, status)
		if results:
			results = []
		if parens or result or results:
			line = prev + line
		else:
			prev = ''


def segmentexport(morphology, functions):
	""" Co-routine that accepts one line at a time.
	Yields tuples (result, status) where ...

	- result is None or a segment delimited by '#BOS ' and '#EOS '
		as a list of lines;
	- status is 1 if the line was consumed, else 0. """
	cur = []
	inblock = False
	line = (yield None, 1)
	while line is not None:
		if line.startswith('#BOS '):
			cur = []
			inblock = True
			line = (yield None, 1)
		elif line.startswith('#EOS '):
			tree, sent = exporttree(cur, morphology)
			handlefunctions(functions, tree)
			line = (yield ((tree, sent), ), 1)
			inblock = False
			cur = []
		elif line.strip():
			if inblock:
				cur.append(line)
			line = (yield None, (1 if inblock else 0))
		else:
			line = (yield None, 0)


def brackettree(treestr, sent, brackets, strtermre):
	""" Parse a single tree presented in bracket format, whether with indices
	or not; sent may be None / empty. """
	if strtermre.search(treestr):  # terminals are not all indices
		treestr = FRONTIERNTRE.sub(' ...)', treestr)
		sent = TERMINALSRE.findall(treestr)
		cnt = count()
		tree = Tree.parse(treestr, brackets=brackets,
				parse_leaf=lambda x: next(cnt))
	else:  # disc. trees with integer indices as terminals
		tree = Tree.parse(treestr, parse_leaf=int,
			brackets=brackets)
		sent = (sent.split() if sent.strip()
				else map(str, range(max(tree.leaves()) + 1)))
	return tree, sent


def exporttree(data, morphology):
	""" Wrapper to get both tree and sentence for tree in export format given
	as list of lines. """
	data = [exportsplit(x) for x in data]
	tree = exportparse(data, morphology)
	sent = []
	for a in data:
		if EXPORTNONTERMINAL.match(a[WORD]):
			break
		sent.append(a[WORD])
	return tree, sent


def treebankfanout(trees):
	""" Get maximal fan-out of a list of trees. """
	# avoid max over empty sequence: 'treebank' may only have unary productions
	try:
		return max((fanout(a), n) for n, tree in enumerate(trees)
				for a in addbitsets(tree).subtrees(lambda x: len(x) > 1))
	except ValueError:
		return 1, 0


if __name__ == '__main__':
	pass



from discodop._grammar cimport ULLong, ULong, UInt, UChar
from libc.string cimport memcpy

cdef extern:
	int __builtin_ffsll (ULLong)
	int __builtin_ctzll (ULLong)
	int __builtin_clzll (ULLong)
	int __builtin_ctzl (ULong)
	int __builtin_popcountl (ULong)
	int __builtin_popcountll (ULLong)

cdef extern from "macros.h":
	int BITSIZE
	int BITSLOT(int b)
	ULong BITMASK(int b)
	ULong TESTBIT(ULong a[], int b)
	void CLEARBIT(ULong a[], int b)

# cpdef functions defined in bit.pyx
#on python integers
cpdef int fanout(arg)
cpdef int pyintnextset(a, int pos)

#on ULLongs
cpdef int bitcount(ULLong vec)
cpdef bint bitminmax(ULLong a, ULLong b)
cpdef bint testbitc(UChar arg, UInt pos)
cpdef bint testbitshort(unsigned short arg, UInt pos)

# cdef inline functions defined here:
#on ULLongs
#cdef inline bint testbit(ULLong vec, UInt pos)
#cdef inline bint testbitint(UInt arg, UInt pos)
#cdef inline int nextset(ULLong vec, UInt pos)
#cdef inline int nextunset(ULLong vec, UInt pos)
#cdef inline int bitlength(ULLong vec)

#on arrays of unsigned long
#cdef inline int abitcount(ULong *vec, short slots)
#cdef inline int anextset(ULong *vec, UInt pos, short slots)
#cdef inline int anextunset(ULong *vec, UInt pos, short slots)
#cdef inline bint subset(ULong *vec1, ULong *vec2, short slots)
#cdef inline void setunion(ULong *dest, ULong *src, short slots)
#cdef inline void ulongset(ULong *dest, ULong value, short slots)
#cdef inline void ulongcpy(ULong *dest, ULong *src, short slots)


# See: http://gcc.gnu.org/onlinedocs/gcc/Other-Builtins.html
cdef inline int nextset(ULLong vec, UInt pos):
	""" Return next set bit starting from pos, -1 if there is none.
	>>> nextset(0b001101, 1)
	2
	"""
	#return (pos + __builtin_ctzll(vec >> pos)) if (vec >> pos) else -1
	# mask instead of shift:
	return __builtin_ffsll(vec & (~0ULL << pos)) - 1
	#ULLong x = vec & ~((1 << pos) - 1)
	#ULLong x = (vec >> pos) << pos
	#return x ? __builtin_ctzll(x) : -1
	#return  __builtin_ffsll(x) - 1


cdef inline int nextunset(ULLong vec, UInt pos):
	""" Return next unset bit starting from pos.
	>> nextunset(0b001101, 2)
	4
	>> nextunset((1<<64)-1, 0)
	64
	"""
	cdef ULLong x = ~vec & (~0ULL << pos)
	return __builtin_ctzll(x) if x else (sizeof(ULLong) * 8)


cdef inline bint testbit(ULLong vec, UInt pos):
	""" Mask a particular bit, return nonzero if set
	>>> testbit(0b0011101, 0)
	True
	>>> testbit(0b0011101, 1)
	False
	>>> testbit(0b100000000000000000000000000000000, 32) != 0
	True
	"""
	return vec & (1ULL << pos) != 0


cdef inline bint testbitint(UInt arg, UInt pos):
	""" Mask a particular bit, return nonzero if set
	>>> testbit(0b0011101, 0)
	1
	>>> testbit(0b0011101, 1)
	0"""
	return arg & (1ULL << pos)


cdef inline int bitlength(ULLong vec):
	""" number of bits needed to represent vector
	(equivalently: index of most significant set bit, plus one)
	>>> bitlength(0b0011101)
	5"""
	return sizeof(vec) * 8 - __builtin_clzll(vec)


cdef inline int abitcount(ULong *vec, short slots):
	""" number of set bits in variable length bitvector """
	cdef short a
	cdef int result = 0
	for a in range(slots):
		result += __builtin_popcountl(vec[a])
	return result


cdef inline int abitlength(ULong *vec, short slots):
	""" number of bits needed to represent vector
	(equivalently: index of most significant set bit, plus one)"""
	cdef short a = slots - 1
	while a and not vec[a]:
		a -= 1
	return (a + 1) * sizeof(ULong) * 8 - __builtin_clzll(vec[a])


cdef inline int anextset(ULong *vec, UInt pos, short slots):
	""" return next set bit starting from pos, -1 if there is none. """
	cdef short a = BITSLOT(pos)
	cdef ULong x
	if a >= slots:
		return -1
	x = vec[a] & (~0UL << (pos % BITSIZE))
	while x == 0UL:
		a += 1
		if a == slots:
			return -1
		x = vec[a]
	return a * BITSIZE + __builtin_ctzl(x)


cdef inline int anextunset(ULong *vec, UInt pos, short slots):
	""" return next unset bit starting from pos. """
	cdef short a = BITSLOT(pos)
	cdef ULong x
	if a >= slots:
		return a * BITSIZE
	x = vec[a] | (BITMASK(pos) - 1)
	while x == ~0UL:
		a += 1
		if a == slots:
			return a * BITSIZE
		x = vec[a]
	return a * BITSIZE + __builtin_ctzl(~x)


cdef inline short iteratesetbits(ULong *vec, short slots,
		ULong *cur, short *idx):
	""" iterate over set bits in an array of unsigned long with 'slots'
	elements. cur and idx are pointers to variables to maintain state, idx
	should be initialized to 0, and cur to the first element of the bit array
	vec. returns the index of a set bit, returns -1 if there are no more set
	bits. result of calling stopped iterator is undefined.
	e.g.,
	ULong vec[4] = {0, 0, 0, 0b10001}, cur = vec[0]
	short idx = 0
	iteratesetbits(vec, 0, 4, &cur, &idx) # returns 0
	iteratesetbits(vec, 0, 4, &cur, &idx) # returns 4
	iteratesetbits(vec, 0, 4, &cur, &idx) # returns -1 """
	cdef short tmp
	while not cur[0]:
		idx[0] += 1
		if idx[0] >= slots:
			return -1
		cur[0] = vec[idx[0]]
	tmp = __builtin_ctzl(cur[0])  # index of bit in current slot
	CLEARBIT(cur, tmp)
	return idx[0] * BITSIZE + tmp


cdef inline void ulongset(ULong *dest, ULong value, short slots):
	""" Like memset, but set one ULong at a time; should be faster
	for small arrays. """
	cdef short a
	for a in range(slots):
		dest[a] = value

#cdef inline void ulongcpy(ULong *dest, ULong *src, short slots):
#	""" Like memcpy, but copy one ULong at a time; should be faster
#	for small arrays. """
#	cdef short a
#	for a in range(slots):
#		dest[a] = src[a]

cdef inline void ulongcpy(ULong *dest, ULong *src, short slots):
	""" memcpy wrapper for unsigned long arrays. """
	memcpy(<char *>dest, <char *>src, slots * sizeof(ULong))


cdef inline void setintersectinplace(ULong *dest, ULong *src, short slots):
	""" dest gets the intersection of dest and src;
	both operands must have at least `slots' slots. """
	cdef short a
	for a in range(slots):
		dest[a] &= src[a]


cdef inline void setunioninplace(ULong *dest, ULong *src, short slots):
	""" dest gets the union of dest and src; both operands must have at least
	`slots' slots. """
	cdef short a
	for a in range(slots):
		dest[a] |= src[a]


cdef inline void setunion(ULong *dest, ULong *src1, ULong *src2, short slots):
	""" dest gets the union of src1 and src2; operands must have at least
	`slots' slots. """
	cdef short a
	for a in range(slots):
		dest[a] = src1[a] | src2[a]


cdef inline bint subset(ULong *vec1, ULong *vec2, short slots):
	""" test whether vec1 is a subset of vec2; i.e., all bits of vec1 should be
	in vec2. """
	cdef short a
	for a in range(slots):
		if (vec1[a] & vec2[a]) != vec1[a]:
			return False
	return True

